from .question_creator_exception import QuestionCreatorException


class DefinitionNotFoundException(QuestionCreatorException):
    pass


class BaitNotFoundException(QuestionCreatorException):
    pass


class MinimalPairNotFoundException(QuestionCreatorException):
    pass


class SpanishBaitsNotFoundException(QuestionCreatorException):
    pass


class ContentNotFoundException(QuestionCreatorException):
    pass


class MaximumLengthExceeded(QuestionCreatorException):
    pass



import typing as t
from dataclasses import dataclass
from typing import Optional

@dataclass
class Content:
    id: str
    phrase: str
    media_type: str
    source: str
    interest_label: str
    translation: Optional[str] = None
    

class TwitterContent(Content):
    image_link: str
    tweet_url: str


class GetyarnContent(Content):
    pass




import random

from ...config import MAX_SYNONYM_ANSWERS
from ...dictionary_factory import pos2words
from ...dictionary_factory import word2multisynonym
from ...exceptions.dictionary_exceptions import WordNotInMultiSynonymDictionary
from ...questions.question import CORRECT
from ...questions.question import Question
from ...questions.question import TEXT
from ..question_creator import QuestionCreator
from question_builder.data import DataQuestion
from typing import List, Dict


class MultiSynonymQuestionCreator(QuestionCreator):

    code = "MSYN"
    baits_code = "pos"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:

        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word
        pos = data_question.pos

        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = self._get_correct_answer(pos, target_lemma)
        question.baits = self._get_baits(pos, question.correct_answer)
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, target_word)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        return question

    def _get_correct_answer(self, pos: str, target_word: str) -> List[str]:
        if (target_word, pos) in word2multisynonym:
            synonym_list = word2multisynonym[(target_word, pos)]
        elif (target_word, None) in word2multisynonym:
            synonym_list = word2multisynonym[(target_word, None)]
        else:
            raise WordNotInMultiSynonymDictionary()

        if len(synonym_list) <= MAX_SYNONYM_ANSWERS:
            synonyms = synonym_list[:]
            random.shuffle(synonyms)
        else:
            synonyms = random.sample(str, MAX_SYNONYM_ANSWERS)
        return synonyms

    def _get_phrase(self, original_phrase: str, target_word: str) -> str:
        return self._underline_word(original_phrase, target_word)

    def _get_baits(self, pos: str, correct_answer: List[str]) -> List[str]:
        baits = set()
        n_baits = MAX_SYNONYM_ANSWERS - len(correct_answer)
        baits.update(random.sample(pos2words[pos], MAX_SYNONYM_ANSWERS))
        while len(baits) < n_baits:
            rnd_pos = random.choice(list(pos2words.keys()))
            if rnd_pos != pos:
                baits.update(pos2words[rnd_pos])
        return list(baits)[:n_baits]

    def _get_options(self, correct_answer: List[str], baits: List[str]) -> List[Dict]:
        """
        Get the baits and the words, and make an array of Options Objects in
        which word is set to True, and all others are False
        """
        options = []
        for word in correct_answer:
            options.append({TEXT: word, CORRECT: True})
        for bait in baits:
            options.append({TEXT: bait, CORRECT: False})
        random.shuffle(options)
        return options



from .antonym_question_creator import AntonymQuestionCreator
from .choice_english_definitions_question_creator import (
    ChoiceEnglishDefinitionsQuestionCreator,
)
from .choice_spanish_definitions_question_creator import (
    ChoiceSpanishDefinitionsQuestionCreator,
)
from .conjugation_question_creator import ConjugationQuestionCreator
from .dual_video_spanish_definitions_question_creator import (
    DualVideoSpanishDefinitionsQuestionCreator,
)
from .fullword_scrabble_question_creator import FullWordScrabbleQuestionCreator
from .fullword_typing_question_creator import FullWordTypingQuestionCreator
from .gap_question_creator import GapQuestionCreator
from .intruder_question_creator import IntruderQuestionCreator
from .means_like_question_creator import MeansLikeQuestionCreator
from .minimal_pairs_question_creator import MinimalPairsQuestionCreator
from .multisynonym_question_creator import MultiSynonymQuestionCreator
from .partword_typing_question_creator import PartWordTypingQuestionCreator
from .relnoun_question_creator import RelNounQuestionCreator
from .sentence_scrabble_question_creator import SentenceScrabbleQuestionCreator
from .subword_question_creator import SubwordQuestionCreator
from .synonym_question_creator import SynonymQuestionCreator
from .tense_verb_choice_question_creator import TenseVerbChoiceQuestionCreator



import random

from question_builder.bp.question_creators.question_creator import QuestionCreator
from question_builder.bp.questions.question import Question
from question_builder.data import DataQuestion
from typing import List

SUBJECT_KEY = "subject"
AUXILIARYVERB_KEY = "conjugated_auxiliaryverb"
TARGET_VERB_CONJUGATIONS_KEY = "lemma_conjugations"
PRESENT_Continuous_KEY = "VBG"
HAD = "had"
HAVE = "have"
WILL = "will"
SAMPLE_N_BAITS = 2


class FuturePerfectChoiceQuestionCreator(QuestionCreator):

    CODE = "FPC"
    BAITS_CODE = "nofpc"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:
        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word

        verbgames_pattern_items = data_question.verbgames_pattern_items
        subject = verbgames_pattern_items[SUBJECT_KEY]
        verbtowill = verbgames_pattern_items[AUXILIARYVERB_KEY]
        target_verb_present_continuous = verbgames_pattern_items[
            TARGET_VERB_CONJUGATIONS_KEY
        ][PRESENT_Continuous_KEY]
        subject_and_verbtowill = self._get_combined_subject_and_verbtowill(
            subject, verbtowill
        )

        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = self._get_correct_answer(
            target_word, subject_and_verbtowill
        )
        question.baits = self._get_baits(
            target_word,
            subject_and_verbtowill,
            target_lemma,
            target_verb_present_continuous,
        )
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, question.correct_answer)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.CODE
        question.baits_type = self.BAITS_CODE
        return question

    @staticmethod
    def _get_combined_subject_and_verbtowill(subject: str, verbtowill: str) -> str:
        if verbtowill == WILL:
            return f"{subject} {verbtowill}"
        else:
            return f"{subject}{verbtowill}"

    @staticmethod
    def _get_correct_answer(target_verb: str, subject_and_verbtowill: str) -> str:
        return f"{subject_and_verbtowill} {HAVE} {target_verb}"

    @staticmethod
    def _get_baits(
        target_verb: str,
        subject_and_verbtowill: str,
        target_verb_lemma: str,
        target_verb_present_continuous: str,
    ) -> List[str]:
        return random.sample(
            [
                f"{subject_and_verbtowill} {HAD} {target_verb}",
                f"{subject_and_verbtowill} {HAVE} {target_verb_present_continuous}",
                f"{subject_and_verbtowill} {HAVE} {target_verb_lemma}",
            ],
            SAMPLE_N_BAITS,
        )

    def _get_phrase(self, original_phrase: str, correct_answer: str) -> str:
        return self._underline_word(original_phrase, correct_answer)



import logging
from collections import defaultdict
from collections import namedtuple
from typing import List, Tuple, Dict

from . import queries
from .. import config
from ..domain.data_converters import to_content
from ..domain.data_converters import to_question
from ..domain.data_converters import to_verb_question
from ..domain.lemma_data import LemmaData
from ..domain.data_question import DataQuestion
from neo4j import DirectDriver
from neo4j import BoltStatementResult


class ContentRepository:
    """
    Handles all the db operations related to user content management
    """

    def __init__(self, driver: DirectDriver):
        """
        Initializes the neo4j db
        """
        self._driver = driver

    def get_lexemas_from_lemma(self, lemma: str) -> List[str]:
        lexemas = []
        with self._driver.session() as session:
            for res in session.read_transaction(self._get_lexemas_from_lemma, lemma):
                lexemas.append(res[0])
        return lexemas

    @staticmethod
    def _get_lexemas_from_lemma(tx, lemma: str) -> BoltStatementResult:
        query = queries.GET_LEXEMAS_FROM_LEMMA
        return tx.run(query, lemma=lemma)

    def get_random_questions_by_pos_without_lexemas(
        self, pos_lemma_list: List[Dict[str, str]], user_id: str, n_questions: int
    ) -> Dict[str, DataQuestion]:
        lemma_to_question = defaultdict()
        with self._driver.session() as session:
            for res in session.read_transaction(
                self._get_random_questions_by_pos_without_lexemas,
                pos_lemma_list,
                user_id,
                n_questions,
            ):
                target_lemma = res[config.WORD_KEY][config.TEXT]
                content_rel_json = res[config.CONTENT_KEY]
                content_rel_json[config.LEMMA_KEY] = target_lemma
                lemma_to_question[target_lemma] = to_question(content_rel_json)
        return lemma_to_question

    @staticmethod
    def _get_random_questions_by_pos_without_lexemas(
        tx, pos_lemma_list: List[Dict[str, str]], user_id: str, n_questions: int
    ) -> BoltStatementResult:
        query = queries.GET_USER_RANDOM_QUESTIONS_WITH_CONTENT_BY_POS_WITHOUT_LEXEMAS
        return tx.run(
            query,
            pos_lemma_list=pos_lemma_list,
            user_id=user_id,
            n_questions=int(n_questions / config.N_INTEREST_TYPES),
        )

    def get_random_questions_by_pos(self, pos_lemma_list: List[Dict[str, str]], user_id: str, n_questions: int) -> Dict[str, DataQuestion]:
        lemma_to_question = defaultdict()
        with self._driver.session() as session:
            for res in session.read_transaction(
                self._get_random_questions_by_pos, pos_lemma_list, user_id, n_questions
            ):
                target_lemma = res[config.WORD_KEY][config.TEXT]
                content_rel_json = res[config.CONTENT_KEY]
                content_rel_json[config.LEMMA_KEY] = target_lemma
                lemma_to_question[target_lemma] = to_question(content_rel_json)
        return lemma_to_question

    @staticmethod
    def _get_random_questions_by_pos(tx, pos_lemma_list: List[Dict[str, str]], user_id: str, n_questions: int) -> BoltStatementResult:
        query = queries.GET_USER_RANDOM_QUESTIONS_WITH_CONTENT_BY_POS
        return tx.run(
            query,
            pos_lemma_list=pos_lemma_list,
            user_id=user_id,
            n_questions=int(n_questions / config.N_INTEREST_TYPES),
        )

    def get_random_questions(
        self, 
        lemma_list: List[str], 
        user_id: str, 
        n_questions: int, 
        n_questions_per_word: int
    ) -> Dict[str, List[LemmaData]]:
        from question_builder.bp.dictionary_factory import word2minimalpair
        lemma_to_data_list = defaultdict(list)
        lemma_list = [{
                        config.LEMMA_KEY: lemma, 
                        config.PAIR_KEY: word2minimalpair.get(lemma, [None])[0]
                    } for lemma in lemma_list]
        with self._driver.session() as session:
            for res in session.read_transaction(
                self._get_random_questions,
                user_id,
                lemma_list,
                n_questions,
                n_questions_per_word,
            ):
                target_lemma = res[config.LEMMA_KEY]
                content_rel_json = res[config.CONTENT_KEY]
                content_rel_json[config.LEMMA_KEY] = target_lemma
                mastered = res[config.MASTERED_KEY]
                level = res[config.LEVEL_KEY]
                mastered = mastered if mastered else False
                level = level if level else 1
                lemma_to_data_list[target_lemma].append(
                    LemmaData(to_question(content_rel_json), level, mastered)
                )
        return lemma_to_data_list

    @staticmethod
    def _get_random_questions(
        tx, user_id: str, 
        lemma_dict_list: List[Dict[str, str]], 
        n_questions: int, 
        n_questions_per_word: int
    ) -> BoltStatementResult:
        query = queries.GET_USER_RANDOM_QUESTIONS_WITH_CONTENT
        result =  tx.run(
            query,
            user_id=user_id,
            lemma_dict_list=lemma_dict_list,
            n_questions=int(n_questions / config.N_INTEREST_TYPES),
            n_questions_per_word=n_questions_per_word,
        )
        return result

    def get_random_questions_paginated(
        self, 
        lemma_list: List[str], 
        user_id: str, 
        page: int, 
        max_items_per_page: int
    ) -> Tuple[int, List[DataQuestion]]:
        question_list = []
        with self._driver.session() as session:
            for res in session.read_transaction(
                self._get_random_questions_paginated,
                lemma_list,
                user_id,
                page,
                max_items_per_page,
            ):
                for content_rel_json in res[config.CONTENT_LIST_KEY]:
                    question_list.append(to_question(content_rel_json))
        return page, question_list

    @staticmethod
    def _get_random_questions_paginated(
        tx, 
        lemma_list: List[int], 
        user_id: str, 
        page: int, 
        max_items_per_page: int
    ) -> BoltStatementResult:
        n_questions = max_items_per_page
        query = queries.GET_USER_RANDOM_QUESTIONS_WITH_CONTENT_PAGINATED
        return tx.run(
            query,
            lemma_list=lemma_list,
            user_id=user_id,
            start=page * n_questions,
            n_questions=n_questions,
        )

    def get_question(self, user_id: str, lemma_list: List[str]) -> DataQuestion:
        with self._driver.session() as session:
            for res in session.read_transaction(
                self._get_random_question, user_id, lemma_list
            ):
                content_rel_json = res[config.Q_VARIABLE_KEY]
                question = to_question(content_rel_json)
                return question

    @staticmethod
    def _get_random_question(tx, user_id: str, lemma_list: str) -> BoltStatementResult:
        query = queries.GET_QUESTION_FROM_USER
        results = tx.run(query, user_id=user_id, lemma_list=lemma_list)
        return results



import re

from question_builder.bp.question_creators.question_creator import QuestionCreator
from question_builder.bp.questions.question import Question
from question_builder.data import DataQuestion
from typing import List

PATTERN_KEY = "pattern"
SUBJECT_KEY = "subject"
LEMMA_CONJUGATIONS = "lemma_conjugations"
VBG = "VBG"
AM = "am"
CONTRACTED_AM = "'m"
IS = "is"
CONTRACTED_IS = "'s"
ARE = "are"
CONTRACTED_ARE = "'re"
INSTRUCTIONS = 'Completa el presente continuo \n Usa el verbo "{}"'

class PresentContinuousFullWordTypingQuestionCreator(QuestionCreator):

    code = "PCFT"
    baits_code = "pcft"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:
        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word

        verbgames_pattern_items = data_question.verbgames_pattern_items
        pattern = verbgames_pattern_items[PATTERN_KEY]
        subject = verbgames_pattern_items[SUBJECT_KEY]

        question = Question()
        question.content_id = content.id
        question.target_lemma = target_lemma
        question.target_word = target_word
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = self._get_correct_answer(pattern, subject)
        question.baits = self._get_baits()
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, pattern, subject)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        question.instructions = INSTRUCTIONS.format(target_lemma)
        return question

    def _get_correct_answer(self, pattern: str, subject: str) -> str:
        correct_answer = pattern.replace(subject, "", 1).replace("'", "").strip()
        correct_answer = correct_answer.replace(CONTRACTED_AM, AM)
        correct_answer = correct_answer.replace(CONTRACTED_IS, IS)
        correct_answer = correct_answer.replace(CONTRACTED_ARE, ARE)
        return correct_answer.replace(" ", "")

    def _get_baits(self) -> List[str]:
        return []

    def _get_pattern_substitution(self, pattern: str, subject: str) -> str:
        final_pattern = pattern
        words_to_remove = pattern.replace(subject, "").replace("'", "").strip()
        for word in words_to_remove.split():
            final_pattern = self._remove_word(final_pattern, word)
        return final_pattern

    def _get_phrase(self, original_phrase: str, pattern: str, subject: str) -> str:
        pattern_substitution = self._get_pattern_substitution(pattern, subject)
        redata = re.compile(re.escape(pattern), re.IGNORECASE)
        phrase = redata.sub(pattern_substitution , original_phrase)
        return phrase



from ...question_getter import QuestionGetter
from .lemma_queries import GET_LEMMA_LIST_DATA
from .present_perfect_continuous_affirmative_queries import GET_N_DATA_QUESTIONS
from .present_perfect_continuous_affirmative_queries import (
    GET_N_DATA_QUESTIONS_PER_WORD,
)


class PRPCCQuestionGetter(QuestionGetter):
    code = "PRPCC"
    query_n_data_questions_per_word = GET_N_DATA_QUESTIONS_PER_WORD
    query_n_data_questions = GET_N_DATA_QUESTIONS
    query_lemma_data = GET_LEMMA_LIST_DATA



import re

from question_builder.bp.question_creators.question_creator import QuestionCreator
from question_builder.bp.questions.question import Question
from question_builder.data import DataQuestion
from typing import List

LEMMA_CONJUGATIONS = "lemma_conjugations"
PATTERN_KEY = "pattern"
VBG = "VBG"
NOT = " not "
NT = "n't"
INSTRUCTIONS = 'Completa la negación'


class PresentContinuousNegativePartWordTypingQuestionCreator(QuestionCreator):

    code = "PCNPT"
    baits_code = "nopcnpt"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:
        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word
        verbgames_pattern_items = data_question.verbgames_pattern_items
        negation = verbgames_pattern_items["negation"]
        pattern = verbgames_pattern_items[PATTERN_KEY]

        question = Question()
        question.content_id = content.id
        question.target_lemma = target_lemma
        question.target_word = target_word
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = negation.replace("'", "").strip()
        question.baits = self._get_baits()
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, pattern, negation)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        question.instructions = INSTRUCTIONS.format(target_lemma)
        return question

    def _get_baits(self) -> List[str]:
        return []

    def _get_pattern_substitution(self, pattern: str, negation: str) -> str:
        final_pattern = pattern
        redata = re.compile(re.escape(negation), re.IGNORECASE)
        if(negation == "n't"):
            final_pattern = redata.sub("_'_", pattern)
        elif(negation == "not"):
            final_pattern = redata.sub("___", pattern)
        return final_pattern

    def _get_phrase(self, original_phrase: str, pattern: str, negation: str) -> str:
        pattern_substitution = self._get_pattern_substitution(pattern, negation)
        redata = re.compile(re.escape(pattern), re.IGNORECASE)
        phrase = redata.sub(pattern_substitution , original_phrase)
        return phrase



GET_USER_RANDOM_QUESTIONS_WITH_CONTENT = """
UNWIND $lemma_dict_list as lemma_dict
MATCH (user:USER{id:$user_id})
MATCH (interest) - [:HAS_LEMMA]-> (lemma:LEMMA)
WHERE (user) - [:LIKES] -> (interest) AND lemma.text = lemma_dict.lemma
OPTIONAL MATCH (user) -[has_seen:HAS_SEEN] -(lemma)
WITH DISTINCT lemma.text AS lemma, lemma_dict.pair as pair, has_seen.level AS level, has_seen.mastered AS mastered

WITH lemma, pair, level, mastered

CALL apoc.cypher.run("

    MATCH (interest) - [:HAS_CONTENT] -> (content:CONTENT) - [has_word:HAS_WORD] -> (word:WORD)-[:HAS_LEMMA]->(:LEMMA{text:{lemma}})       
    WHERE (:USER{id:{user_id}}) - [:LIKES] -> (interest)
    WITH interest, content, has_word, word, RAND() AS rnd LIMIT {n_questions}

    WITH interest, content, has_word, word ORDER BY rnd LIMIT 1
    MATCH (interest) - [:HAS_CONTENT] -> (bait_content:CONTENT) - [has_word_:HAS_WORD] -> (word_:WORD)
    WHERE has_word_.lemma <> {lemma}
	
    WITH interest, content, bait_content, has_word, word LIMIT 1
	OPTIONAL MATCH (pair_word:WORD{text:{pair}})
    OPTIONAL MATCH (interest) - [:HAS_CONTENT] -> (pair_content:CONTENT) - [:HAS_WORD] -> (pair_word)
    
    RETURN {
            content: {
                id: content.id,
                translation: content.translation,
                source: content.source,
                media_type: content.media_type,
                phrase: content.phrase,
                interest_label: labels(interest)[0]
            }, 
            bait_content: {
                id: bait_content.id,
                translation: bait_content.translation,
                source: bait_content.source,
                media_type: bait_content.media_type,
                phrase: bait_content.phrase,
                interest_label: labels(interest)[0]
            }, 
            pair_content: {
                id: pair_content.id,
                translation: pair_content.translation,
                source: pair_content.source,
                media_type: pair_content.media_type,
                phrase: pair_content.phrase,
                interest_label: labels(interest)[0]
            },  
            lemma:{lemma}, 
            interest: interest.name, 
            word: word.text, 
            pos: has_word.pos
            } as content LIMIT 1

    ", {user_id: $user_id, lemma: lemma, pair: pair, n_questions: $n_questions, n_questions_per_word: $n_questions_per_word}) YIELD value
RETURN lemma, value.content AS content, level, mastered
"""


GET_USER_RANDOM_QUESTIONS_WITH_CONTENT_PAGINATED = """
MATCH (interest) - [:HAS_CONTENT] -> (content:CONTENT) - [has_word:HAS_WORD] -> (word:WORD) -[:HAS_LEMMA]-> (lemma_:LEMMA)
WHERE (:USER{id:{user_id}}) - [:LIKES] -> (interest) and lemma_.text in $lemma_list

WITH interest, content, has_word, word, lemma_, RAND() AS rnd SKIP {start} LIMIT {n_questions}

WITH {
        content: {
            id: content.id,
            translation: content.translation,
            source: content.source,
            media_type: content.media_type,
            phrase: content.phrase,
            interest_label: labels(interest)[0]
        }, 
        bait_content: null, 
        pair_content: null, 
        lemma:lemma_.text, 
        interest: interest.name, 
        word: word.text, 
        pos: has_word.pos
    } AS content ORDER BY rnd

RETURN content.lemma as lemma, collect(content) AS content_list"""


GET_QUESTION_FROM_USER = """
UNWIND $lemma_list AS lemma

CALL apoc.cypher.run("

    MATCH (user:USER{id:{user_id}})
	MATCH (interest) - [:HAS_CONTENT] -> (content:CONTENT) - [has_word:HAS_WORD] -> (word_:WORD) - [:HAS_LEMMA] -> (lemma_:LEMMA{text:{lemma}})
	WHERE (user) -[:LIKES] -> (interest)
	RETURN {
        content: {
            id: content.id,
            translation: content.translation,
            source: content.source,
            media_type: content.media_type,
            phrase: content.phrase,
            interest_label: labels(interest)[0]
        }, 
        bait_content: null, 
        pair_content: null, 
        lemma:lemma_.text,
        interest:interest.name,
        word:word_.text,
        pos:has_word.pos
    } AS content LIMIT 10

    ", {lemma:lemma, user_id: $user_id}) YIELD value
RETURN value.content AS qu LIMIT 1
"""


GET_USER_RANDOM_QUESTIONS_WITH_CONTENT_BY_POS = """
UNWIND $pos_lemma_list AS pos_lemma
CALL apoc.cypher.run("

    CALL apoc.cypher.run('

        MATCH (interest) - [:HAS_CONTENT] -> (content:CONTENT) - [has_word:HAS_WORD{pos:{pos_lemma}.pos}] -> (word:WORD) -[:HAS_LEMMA]-(lemma:LEMMA{text:{pos_lemma}.text})
        WHERE (:USER{id:{user_id}}) - [:LIKES] ->(interest)
        RETURN {
            content: {
                id: content.id,
                translation: content.translation,
                source: content.source,
                media_type: content.media_type,
                phrase: content.phrase,
                interest_label: labels(interest)[0]
            },
            interest:interest.name,
            word:word.text,
            pos:has_word.pos
        } AS content LIMIT {n_questions}

    ', {pos_lemma:{pos_lemma}, user_id:{user_id}, n_questions:{n_questions}}) YIELD value
    WITH pos_lemma, value.content AS content, RAND() AS rnd

    RETURN pos_lemma, content, rnd ORDER BY rnd LIMIT 1

    ", {pos_lemma: pos_lemma, user_id: $user_id, n_questions: $n_questions}) YIELD value
RETURN value.pos_lemma AS word, value.content AS content
"""


GET_USER_RANDOM_QUESTIONS_WITH_CONTENT_BY_POS_WITHOUT_LEXEMAS = """
UNWIND $pos_lemma_list AS pos_lemma
CALL apoc.cypher.run("

    CALL apoc.cypher.run('

        MATCH (interest) - [:HAS_CONTENT] -> (content:CONTENT) - [has_word:HAS_WORD{pos:{pos_lemma}.pos}] -> (word:WORD{text:{pos_lemma}.text})
        WHERE (:USER{id:{user_id}}) - [:LIKES] ->(interest)
        RETURN {
            content: {
                id: content.id,
                translation: content.translation,
                source: content.source,
                media_type: content.media_type,
                phrase: content.phrase,
                interest_label: labels(interest)[0]
            },
            interest:interest.name,
            word:word.text,
            pos:has_word.pos
        } AS content LIMIT {n_questions}

    ', {pos_lemma:pos_lemma, user_id:{user_id}, n_questions:{n_questions}}) YIELD value
    WITH pos_lemma, value.content AS content, RAND() AS rnd

    RETURN pos_lemma, content, rnd ORDER BY rnd LIMIT 1

    ", {pos_lemma: pos_lemma, user_id: $user_id, n_questions: $n_questions}) YIELD value
RETURN value.pos_lemma AS word, value.content AS content
"""


GET_LEXEMAS_FROM_LEMMA = """
MATCH (word:WORD) -[:HAS_LEMMA] -(lemma:LEMMA{text:$lemma})
WHERE word.text <> lemma.text
RETURN word.text
"""



import random

from ...dictionary_factory import word2minimalpair
from ...exceptions.dual_video_exceptions import ContentNotFoundException
from ...exceptions.dual_video_exceptions import MinimalPairNotFoundException
from ...questions.dual_video_question import DualVideoQuestion
from ..dual_video_question_creator import DualVideoQuestionCreator
from ..question_creator import QuestionCreator
from question_builder.data import DataQuestion
from question_builder.data import Content


class MinimalPairsQuestionCreator(DualVideoQuestionCreator):

    code = "MP"
    baits_code = "nobaits"

    def create(self, data_question: DataQuestion, user_id: str) -> DualVideoQuestion:
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word
        incorrect_content = self._get_bait_content(target_lemma, data_question)
        correct_content = data_question.content
        question = DualVideoQuestion()
        question.content_id = correct_content.id
        question.target_lemma = target_lemma
        question.target_word = target_word
        medias_list = self._get_medias([correct_content, incorrect_content])
        question.links = medias_list.links
        question.media_types = medias_list.media_types
        question.phrases = medias_list.phrases
        question.correct_values = medias_list.correct_values
        question.options = None
        question.phrase = self._get_phrase(correct_content)
        question.original_phrase = correct_content.phrase
        question.phrase_translation = self._get_translation(correct_content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        return question

    def _get_bait_content(self, target_lemma: str, data_question: DataQuestion) -> Content:
        if target_lemma in word2minimalpair:
            content = data_question.pair_content
            if content:
                return content
            raise ContentNotFoundException()
        raise MinimalPairNotFoundException()

    def _get_phrase(self, content) -> str:
        return content.phrase

    def _get_translation(self, content) -> str:
        if hasattr(content, "translation"):
            if content.translation is not None:
                phrase_translation = content.translation.strip()
            else:
                phrase_translation = ""
        else:
            phrase_translation = ""
        return phrase_translation



import random
import re

from ...dictionary_factory import BAITS
from ...dictionary_factory import CORRECT as DICT_CORRECT
from ...dictionary_factory import word2presuffix
from ...exceptions.dictionary_exceptions import WordNotInPreSuffixDictionary
from ...exceptions.pos_exceptions import WordHasNotValidPos
from ...questions.question import CORRECT
from ...questions.question import N_BAITS
from ...questions.question import Question
from ...questions.question import TEXT
from ...validators import pos_validators
from ..question_creator import QuestionCreator
from question_builder.data import DataQuestion
from typing import List


class SubwordQuestionCreator(QuestionCreator):

    code = "SWS"
    baits_code = "nopreffixsuffix"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:
        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word
        pos = data_question.pos

        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = self._get_correct_answer(pos, target_word)
        question.baits = self._get_baits(target_word)
        question.options = self._get_options(
            question.correct_answer, question.baits, target_word
        )
        question.phrase = self._get_phrase(
            content.phrase, target_word, question.correct_answer
        )
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        return question

    def _get_correct_answer(self, pos: str, target_word: str) -> str:
        valid_pos = (
            pos_validators.is_noun(pos)
            or pos_validators.is_verb(pos)
            or pos_validators.is_adverb(pos)
        )
        if not valid_pos:
            raise WordHasNotValidPos()
        if target_word not in word2presuffix:
            raise WordNotInPreSuffixDictionary()
        return word2presuffix[target_word][DICT_CORRECT]

    def _get_options(self, correct_answer: str, baits: List[str], target_word: str) -> List[str]:
        """
        Get the baits and the words, and make an array of Options Objects in
        which word is set to True, and all others are False
        """
        is_prefix = target_word.startswith(correct_answer)
        options = [
            {TEXT: self._add_line_to_option(correct_answer, is_prefix), CORRECT: True}
        ]
        for bait in baits:
            options.append(
                {TEXT: self._add_line_to_option(bait, is_prefix), CORRECT: False}
            )
        random.shuffle(options)
        return options

    def _get_phrase(self, original_phrase: str, target_word: str, correct_answer: str) -> str:
        return self._remove_word(original_phrase, target_word, correct_answer)

    def _get_baits(self, target_word: str) -> List[str]:
        return random.sample(word2presuffix[target_word][BAITS], N_BAITS)

    def _add_line_to_option(self, option: str, is_prefix: bool) -> str:
        if is_prefix:
            return option + "-"
        else:
            return "-" + option

    def _remove_word(self, phrase: str, target_word: str, correct_answer: str) -> str:
        is_prefix = target_word.startswith(correct_answer)
        substitution = target_word.replace(
            correct_answer, self._add_space_to_dash(is_prefix), 1
        )
        phrase = re.sub(
            r"\b" + target_word + r"(\b|(?=n't))",
            substitution,
            phrase,
            flags=re.I,
            count=1,
        )
        phrase = self._clean_phrase(phrase)
        phrase = self._underline_word(phrase, substitution)
        return phrase

    def _add_space_to_dash(self, is_prefix: bool) -> str:
        if is_prefix:
            return "    -"
        else:
            return "-    "



from ..config import QuestionMode
from neo4j import DirectDriver
from neo4j import BoltStatementResult
from typing import List
from typing import Dict
from question_builder.data.domain.lemma_data import LemmaData
from .question_getter import QuestionGetter


class QuestionTypeRepository:
    """
    Handles all the db operations related to user content management
    """

    def __init__(self, driver: DirectDriver):
        """
        Initializes the neo4j db
        """
        self._driver = driver
        self._registry = {}

    def get_random_questions(
        self,
        code: str,
        user_id: str,
        lemmas: List[str],
        mode: str = QuestionMode.N_QUESTIONS_PER_WORD_MODE,
        n_questions: int = None,
        n_questions_per_word: int = None,
    ) -> Dict[str, List[LemmaData]]:
        question_getter = self._registry[code]

        return question_getter(self._driver, mode).get(
            user_id, lemmas, n_questions, n_questions_per_word
        )

    def register_question_getter(self, code: str, question_getter: QuestionGetter):
        self._registry[code] = question_getter



import random
import re

from question_builder.bp.question_creators.question_creator import QuestionCreator
from question_builder.bp.questions.question import Question
from question_builder.data import DataQuestion
from typing import List

LEMMA_CONJUGATIONS = "lemma_conjugations"
AUXILIARY_VERB = "conjugated_auxiliaryverb"
PATTERN_KEY = "pattern"
VBG = "VBG"
BE = "be"
AM = "am"
CONTRACTED_AM = "'m"
IS = "is"
CONTRACTED_IS = "'s"
ARE = "are"
CONTRACTED_ARE = "'re"
INSTRUCTIONS = 'Completa el presente continuo \n Usa el verbo "{}"'


class PresentContinuousPartWordTypingQuestionCreator(QuestionCreator):

    code = "PCPT"
    baits_code = "nopcpt"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:
        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word

        verbgames_pattern_items = data_question.verbgames_pattern_items
        pattern = verbgames_pattern_items[PATTERN_KEY]
        auxiliary_verb = verbgames_pattern_items[AUXILIARY_VERB]

        question = Question()
        question.content_id = content.id
        question.target_lemma = target_lemma
        question.target_word = target_word
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = self._get_correct_answer(
            target_word, auxiliary_verb
        )
        question.baits = self._get_baits()
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, pattern, question.correct_answer, auxiliary_verb)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        question.instructions = INSTRUCTIONS.format(self._get_correct_answer_lemma(question.correct_answer, target_lemma))
        return question

    def _get_correct_answer(self, target_word: str, auxiliary_verb: str) -> str:
        if bool(random.getrandbits(1)):
            return target_word
        return auxiliary_verb.replace("'", "").strip()

    def _get_baits(self) -> List[str]:
        return []

    def _get_correct_answer_lemma(self, correct_answer: str, target_lemma: str) -> str:
        if(correct_answer in [AM, IS, ARE, CONTRACTED_AM, CONTRACTED_IS, CONTRACTED_ARE]):
            return BE
        else:
            return target_lemma


    def _get_pattern_substitution(self, pattern: str, correct_answer: str, auxiliary_verb: str) -> str:
        if(auxiliary_verb.replace("'", "") == correct_answer and \
            auxiliary_verb.startswith("'")):
            redata = re.compile(re.escape(auxiliary_verb), re.IGNORECASE)  
            final_pattern = redata.sub("'"+len(correct_answer)*"_", pattern) 
            return final_pattern

        redata = re.compile(re.escape(correct_answer), re.IGNORECASE)   
        final_pattern = redata.sub(len(correct_answer)*"_", pattern)
        return final_pattern

    def _get_phrase(self, original_phrase: str, pattern: str, correct_answer: str, auxiliary_verb: str) -> str:
        pattern_substitution = self._get_pattern_substitution(pattern, 
                                                            correct_answer, 
                                                            auxiliary_verb)
        redata = re.compile(re.escape(pattern), re.IGNORECASE)
        phrase = redata.sub(pattern_substitution , original_phrase)
        return phrase



from question_builder.bp.exceptions.dual_video_exceptions import (
    ContentNotFoundException,
)
from question_builder.bp.question_creators.dual_video_question_creator import (
    DualVideoQuestionCreator,
)
from question_builder.bp.questions.dual_video_question import DualVideoQuestion
from question_builder.data import DataQuestion
from question_builder.data import Content


class FutureDualVideoQuestionCreator(DualVideoQuestionCreator):

    code = "FDV"
    baits_code = "nofdv"

    def create(self, data_question: DataQuestion, user_id: str) -> DualVideoQuestion:

        target_lemma = data_question.target_lemma
        target_word = data_question.target_word
        incorrect_content = self._get_bait_content(data_question.bait_content)
        correct_content = data_question.content

        question = DualVideoQuestion()
        question.content_id = correct_content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        medias_list = self._get_medias([correct_content, incorrect_content])
        question.links = medias_list.links
        question.media_types = medias_list.media_types
        question.phrases = medias_list.phrases
        question.correct_values = medias_list.correct_values
        question.options = None
        question.phrase = self._get_phrase(correct_content)
        question.original_phrase = correct_content.phrase
        question.phrase_translation = self._get_translation(correct_content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        return question

    def _get_bait_content(self, content: Content) -> Content:
        if content:
            return content
        raise ContentNotFoundException()

    def _get_phrase(self, content: Content) -> str:
        return content.phrase

    def _get_translation(self, content: Content) -> str:
        if hasattr(content, "translation"):
            if content.translation is not None:
                phrase_translation = content.translation.strip()
            else:
                phrase_translation = ""
        else:
            phrase_translation = ""
        return phrase_translation



import random
import re

from ...exceptions.pos_exceptions import InvalidMorpheme
from ...exceptions.pos_exceptions import NotConjugation
from ...exceptions.pos_exceptions import WordNotVerb
from ...questions.question import Question
from ...validators import pos_validators
from ..question_creator import QuestionCreator
from question_builder.data import DataQuestion
from typing import List

MAX_OPTIONS = 10
COMMON_BAITS = "aeioungrmd"


class ConjugationQuestionCreator(QuestionCreator):

    code = "CJG"
    baits_code = "nobaits"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:

        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word
        pos = data_question.pos

        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = self._get_correct_answer(
            pos, target_word, target_lemma
        )
        question.baits = self._get_baits(question.correct_answer)
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(
            content.phrase, target_word, question.correct_answer
        )
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        return question

    def _get_correct_answer(self, pos: str, target_word: str, target_lemma: str) -> str:
        if not pos_validators.is_verb(pos):
            raise WordNotVerb()
        if target_word == target_lemma:
            raise NotConjugation()
        morpheme = self._get_morpheme(target_lemma, target_word)
        if morpheme == target_word:
            raise InvalidMorpheme()
        return morpheme

    def _get_baits(self, correct_answer: str) -> List[str]:
        remaining = MAX_OPTIONS - len(correct_answer)
        baits = self._get_random_baits(remaining)
        return ["".join(baits)]

    def _get_random_baits(self, n: int) -> List[str]:
        return random.sample(COMMON_BAITS, n)

    def _get_phrase(self, original_phrase: str, target_word: str, correct_answer: str) -> str:
        return self._remove_word(original_phrase, target_word, correct_answer)

    def _remove_word(self, phrase: str, target_word: str, correct_answer: str) -> str:
        # speaking, ing
        substitution = re.sub("%s$" % correct_answer, "-    ", target_word)
        phrase = re.sub(
            r"\b" + target_word + r"(\b|(?=n't))",
            substitution,
            phrase,
            flags=re.I,
            count=1,
        )
        phrase = self._clean_phrase(phrase)
        phrase = self._underline_word(phrase, substitution)
        return phrase

    def _get_morpheme(self, target_lemma: str, target_word: str) -> str:
        for i, (char1, char2) in enumerate(zip(target_lemma, target_word)):
            if char1 != char2:
                return target_word[i:]

        return target_word[i + 1 :]  # noqa: E203



GET_N_DATA_QUESTIONS_PER_WORD = """
UNWIND $lemma_list as lemma

CALL apoc.cypher.run("

    MATCH (interest) - [:HAS_CONTENT] -> (content:CONTENT) - [pattern_rel:HAS_PATTERN{lemma:{lemma}}] -> (:PATTERN{code:'PACC'})
    MATCH (content)-[:HAS_WORD{position: pattern_rel.position - 2}]-(subject:WORD)
    MATCH (content)-[:HAS_WORD{position: pattern_rel.position - 1}]-(auxiliaryverb:WORD)
    MATCH (content)-[rel_target_word:HAS_WORD{position : pattern_rel.position}]-(target_word:WORD)
    WHERE (:USER{id:{user_id}}) - [:LIKES] ->(interest)
    WITH {
        lemma: pattern_rel.lemma,
        content: {
            id: content.id,
            translation: content.translation,
            source: content.source,
            media_type: content.media_type,
            phrase: content.phrase,
            interest_label: labels(interest)[0]
        },
        interest:interest.name,
        word:target_word.text,
        pos:rel_target_word.pos,
        verbgames_pattern_items:{
                                subject: subject.text,
                                pattern: pattern_rel.pattern,
                                conjugated_auxiliaryverb: auxiliaryverb.text
                                }
    } AS content, RAND() AS rnd LIMIT 10

RETURN content LIMIT {n_questions_per_word}

", {user_id: $user_id, lemma: lemma, n_questions_per_word: $n_questions_per_word}) YIELD value

RETURN value.content as content
"""

GET_N_DATA_QUESTIONS = """
MATCH (interest) - [:HAS_CONTENT] -> (content:CONTENT) - [pattern_rel:HAS_PATTERN] -> (:PATTERN{code:'PACC'})
MATCH (content)-[:HAS_WORD{position: pattern_rel.position - 2}]-(subject:WORD)
MATCH (content)-[:HAS_WORD{position: pattern_rel.position - 1}]-(auxiliaryverb:WORD)
MATCH (content)-[rel_target_word:HAS_WORD{position: pattern_rel.position}]-(target_word:WORD)
WHERE (:USER{id:{user_id}}) - [:LIKES] ->(interest) AND pattern_rel.lemma in {lemma_list}
WITH {
    lemma:pattern_rel.lemma,
    content: {
        id: content.id,
        translation: content.translation,
        source: content.source,
        media_type: content.media_type,
        phrase: content.phrase,
        interest_label: labels(interest)[0]
    },
	interest:interest.name,
	word:target_word.text,
	pos:rel_target_word.pos,
	verbgames_pattern_items:{
							subject: subject.text,
							pattern: pattern_rel.pattern,
							conjugated_auxiliaryverb: auxiliaryverb.text
							}
} AS content, RAND() AS rnd ORDER BY rnd LIMIT 50

RETURN content LIMIT $n_questions
"""



from .question_creator_exception import QuestionCreatorException


class WordNotAdjective(QuestionCreatorException):
    pass


class WordNotVerb(QuestionCreatorException):
    pass


class NotConjugation(QuestionCreatorException):
    pass


class InvalidMorpheme(QuestionCreatorException):
    pass


class WordNotNoun(QuestionCreatorException):
    pass


class WordHasNotValidPos(QuestionCreatorException):
    pass



import random

from question_builder.bp.question_creators.question_creator import QuestionCreator
from question_builder.bp.questions.question import Question
from question_builder.data import DataQuestion
from typing import List

SUBJECT_KEY = "subject"
NEGATION_ADVERB_KEY = "negation"
NOT = "not"
NO = "no"
DONT = "don't"
WILL = "will"
WONT = "won't"
SAMPLE_N_BAITS = 2


class SimpleFutureNegationChoiceQuestionCreator(QuestionCreator):

    CODE = "SFNC"
    BAITS_CODE = "nosfnc"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:
        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word

        verbgames_pattern_items = data_question.verbgames_pattern_items
        subject = verbgames_pattern_items[SUBJECT_KEY]
        negation_adverb = verbgames_pattern_items[NEGATION_ADVERB_KEY]
        verbtowill_and_negation_adverb = self._get_combined_verbtowill_and_negation_adverb(
            negation_adverb
        )

        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = self._get_correct_answer(
            target_word, subject, verbtowill_and_negation_adverb
        )
        question.baits = self._get_baits(
            target_word, subject, verbtowill_and_negation_adverb
        )
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, question.correct_answer)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.CODE
        question.baits_type = self.BAITS_CODE
        return question

    @staticmethod
    def _get_combined_verbtowill_and_negation_adverb(negation_adverb: str) -> str:
        if negation_adverb == NOT:
            return f"{WILL} {NOT}"
        else:
            return WONT

    @staticmethod
    def _get_correct_answer(target_verb: str, subject: str, verbtowill_and_negation_adverb: str) -> str:
        return f"{subject} {verbtowill_and_negation_adverb} {target_verb}"

    @staticmethod
    def _get_baits(target_verb: str, subject: str, verbtowill_and_negation_adverb: str) -> List[str]:
        return random.sample(
            [
                f"{subject} {DONT} {WILL} {target_verb}",
                f"{subject} {WILL} {NO} {target_verb}",
                f"{subject} {NOT} {WILL} {target_verb}",
            ],
            SAMPLE_N_BAITS,
        )

    def _get_phrase(self, original_phrase: str, correct_answer: str) -> str:
        return self._underline_word(original_phrase, correct_answer)



import logging
import random

from .exceptions.missing_params_exceptions import MissingParamsException
from .exceptions.question_creator_exception import QuestionCreatorException
from question_builder.data import DataQuestion
from question_builder.data import ContentRepository
from .question_creators.question_creator import QuestionCreator
from .questions.question import Question
from .question_pack import CreatorNode
from .question_pack import QuestionPack
from typing import Dict, List

class QuestionBuilder:
    """
    Builds a question given the raw data DataQuestion information.
    It uses the responsibility chain design pattern to find the question
    creator that takes charge of creating the bp question.
    """

    def __init__(self, 
                content_repository: ContentRepository, 
                question_type_to_creator: Dict[str, QuestionCreator] = None):
        self.content_repository = content_repository
        self.question_type_to_creator = question_type_to_creator
        self.question_pack_registry: Dict[str, QuestionPack] = {}

    def create(
        self,
        data_question: DataQuestion,
        user_id: str,
        question_pack_code: str = None,
        question_type: str = None,
        level: int = None,
        mastered: bool = False,
        force_type: bool = False,
    ) -> Question:
        """
        Creates a question according to the question pack given in the
        arguments
        """
        logging.debug(
            "\nQuestion builder: {} {} {} {} {} {}".format(
                data_question,
                user_id,
                question_pack_code,
                question_type,
                level,
                mastered,
            )
        )

        if force_type:
            if not self.question_type_to_creator:
                raise MissingParamsException(
                    "question_type_to_creator dictionary is required"
                )
            if not question_type:
                raise MissingParamsException("question_type parameter is required")

            return self._get_question_by_type(data_question, user_id, question_type)

        if not self.question_pack_registry:
            raise MissingParamsException(
                "question_pack_registry dictionary is required"
            )

        if not question_pack_code:
            raise MissingParamsException("question_pack_code parameter is required")

        question_pack = self.question_pack_registry[question_pack_code]
        question_type_code = self._get_question_type_code(
            question_type,
            level,
            question_pack.question_types,
            question_pack.level_to_question_types,
        )
        creator_node = question_pack.responsibility_chain[question_type_code]
        question_ = self._get_question(
            creator_node, data_question, self.content_repository, user_id
        )
        question_.mastered = mastered
        if level:
            question_.level = level
            question_.level_to_master = question_pack.level_to_master
        return question_

    def _get_question_type_code(
        self,
        question_type: str,
        level: int,
        default_question_types: List[str],
        default_level_to_question_types: Dict[int, List[str]],
    ) -> str:
        """
        Returns the same question type if given in the argument, else return a
        default one
        """
        if question_type is None:
            if level is not None:
                max_question_level = max(list(default_level_to_question_types.keys()))
                if level > max_question_level:
                    level = max_question_level
                question_type = default_level_to_question_types[level][0]
            else:
                question_type = random.choice(default_question_types)
        return question_type

    def _get_question(
        self, 
        question_creator_node: CreatorNode, 
        data_question: DataQuestion, 
        content_repository: ContentRepository, 
        user_id: str
    ) -> Question:
        """
        Go down the responsibility chain until we find a question that goes
        along with the data_question
        """
        while question_creator_node:
            try:
                question_creator_class = question_creator_node.question_creator
                question_creator = question_creator_class(content_repository)
                question = question_creator.create(data_question, user_id)
                return question
            except QuestionCreatorException:
                question_creator_node = question_creator_node.next

    def _get_question_by_type(self, data_question: DataQuestion, user_id: str, question_type: str) -> Question:
        question_creator_class = self.question_type_to_creator[question_type]
        question_creator = question_creator_class(self.content_repository)
        return question_creator.create(data_question, user_id)

    def register_question_pack(self, code: str, question_pack: QuestionPack):
        """
        Registers question_pack
        """
        self.question_pack_registry[code] = question_pack



import random
import re

from question_builder.bp.question_creators.question_creator import QuestionCreator
from question_builder.bp.questions.question import Question
from question_builder.data import DataQuestion
from typing import List

PATTERN_KEY = "pattern"
SUBJECT_KEY = "subject"
AUXILIARYVERB_KEY = "conjugated_auxiliaryverb"
LEMMA_CONJUGATIONS = "lemma_conjugations"
VBG = "VBG"
VBD = "VBD"
AM = "am"
AAM = "'m"
IS = "is"
AIS = "'s"
ARE = "are"
AARE = "'re"
WAS = "was"
WERE = "were"
AINT = "ain't"
N_BAITS = 2
INSTRUCTIONS = 'Elige la opción de presente continuo'


class PresentContinuousChoiceQuestionCreator(QuestionCreator):

    code = "PCC"
    baits_code = "nopcc"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:

        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word

        verbgames_pattern_items = data_question.verbgames_pattern_items
        pattern = verbgames_pattern_items[PATTERN_KEY]
        subject = verbgames_pattern_items[SUBJECT_KEY]
        lemma_conjugations = verbgames_pattern_items[LEMMA_CONJUGATIONS]
        verb_past = lemma_conjugations[VBD]
        verb_present_continuous = lemma_conjugations[VBG]
        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = pattern
        question.baits = self._get_baits(
            subject, pattern, target_lemma, verb_past, verb_present_continuous
        )
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, pattern)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        question.instructions = INSTRUCTIONS.format(target_lemma)
        return question

    def _get_baits(
        self, subject: str, pattern: str, target_lemma: str, verb_past: str, verb_present_continuous: str
    ) -> List[str]:
        verb_to_be = self._get_verb_to_be(pattern)
        past_verb_to_be = self._get_past_verb_to_be(verb_to_be)
        return random.sample(
            [
                f"{subject}{past_verb_to_be} {verb_present_continuous}",
                f"{subject}{verb_to_be} {verb_past}",
            ],
            2,
        )

    def _get_phrase(self, original_phrase: str, pattern: str) -> str:
        redata = re.compile(re.escape(pattern), re.IGNORECASE)
        phrase = redata.sub("_" * len(pattern), original_phrase)
        return phrase

    def _get_verb_to_be(self, pattern: str) -> str:
        if f" {AM} " in pattern:
            return f" {AM}"
        elif f"{AAM} " in pattern:
            return f"{AAM}"
        elif f" {IS} " in pattern:
            return f" {IS}"
        elif f"{AIS} " in pattern:
            return f"{AIS}"
        elif f" {ARE} " in pattern:
            return f" {ARE}"
        elif f"{AARE} " in pattern:
            return f"{AARE}"
        elif f"{AINT} " in pattern:
            return f" {AINT}"

    def _get_past_verb_to_be(self, verb: str) -> str:
        if f"{AM}" in verb or f"{AAM}" in verb:
            return f" {WAS}"
        elif f"{IS}" in verb or f"{AIS}" in verb:
            return f" {WAS}"
        elif f"{ARE}" in verb or f"{AARE}" in verb:
            return f" {WERE}"
        if f"{AINT}" in verb:
            return f" {WAS}"


from typing import Optional, List, Dict

ID = "id"
MEDIA = "media"
MEDIA_TYPE = "mediaType"
MEDIA_LINK = "link"
PHRASE = "phrase"
ORIGINAL_PHRASE = "originalphrase"
TRANSLATION_KEY = "phrase_translation"
QUESTION_TYPE = "questiontype"
OPTIONS = "options"
LEVEL = "level"
IS_LAST_LEVEL = "lastlevel"
TARGET_LEMMA = "targetlemma"
MASTERED = "mastered"
CORRECT = "correct"
N_BAITS = 3
GETYARN_KEY = "getyarn"
TWITTER_KEY = "twitter"
VIDEO = "video"
IMAGE = "image"
TEXT = "text"
INSTRUCTIONS = "instructions"


class Question:
    def __init__(self):
        self.content_id: Optional[str] = None
        self.target_lemma: Optional[str] = None
        self.links: Optional[List[str]] = None
        self.media_types: Optional[List[str]] = None
        self.options: Optional[List[Dict]] = None
        self.phrase: Optional[str] = None
        self.original_phrase: Optional[str] = None
        self.phrase_translation: Optional[str] = None
        self.question_type: Optional[str] = None
        self.baits_type: Optional[str] = None
        self.level_to_master: Optional[int] = 1
        self.level: Optional[int] = 1
        self.mastered: Optional[bool] = False
        self.instructions: Optional[str] = ""

    def __repr__(self):
        return "Question({})".format(self.id)

    @property
    def id(self) -> str:
        return "{}_{}_{}_{}_{}".format(
            self.content_id,
            self.target_lemma,
            self.target_word,
            self.question_type,
            self.baits_type,
        )

    @property
    def is_last_level(self) -> bool:
        return self.level >= self.level_to_master

    def to_json(self) -> Dict:
        question = {
            ID: self.id,
            MEDIA: [
                {MEDIA_TYPE: media_type, MEDIA_LINK: link}
                for media_type, link in zip(self.media_types, self.links)
            ],
            PHRASE: self.phrase,
            ORIGINAL_PHRASE: self.original_phrase,
            TRANSLATION_KEY: self.phrase_translation,
            OPTIONS: self.options,
            QUESTION_TYPE: self.question_type,
            LEVEL: self.level,
            IS_LAST_LEVEL: self.is_last_level,
            TARGET_LEMMA: self.target_lemma,
            MASTERED: self.mastered,
            INSTRUCTIONS: self.instructions
        }
        return question



from ...question_getter import QuestionGetter
from .lemma_queries import GET_LEMMA_LIST_DATA
from .simple_present_affirmative_queries import GET_N_DATA_QUESTIONS
from .simple_present_affirmative_queries import GET_N_DATA_QUESTIONS_PER_WORD


class STCQuestionGetter(QuestionGetter):
    code = "STC"
    query_n_data_questions_per_word = GET_N_DATA_QUESTIONS_PER_WORD
    query_n_data_questions = GET_N_DATA_QUESTIONS
    query_lemma_data = GET_LEMMA_LIST_DATA



import random

import editdistance

from ...exceptions.lexemas_exceptions import NoLexemasFound
from ...exceptions.pos_exceptions import WordNotVerb
from ...questions.question import Question
from ...validators import pos_validators
from ..question_creator import QuestionCreator
from question_builder.data import DataQuestion
from typing import List


class TenseVerbChoiceQuestionCreator(QuestionCreator):
    code = "TVC"
    baits_code = "nobaits"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:
        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word
        pos = data_question.pos
        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = target_word
        question.baits = self._get_baits(target_lemma, pos)
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, target_word)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        return question

    def _choose_bait(self, possible_baits: List[str], target_lemma: str) -> List[str]:
        distances = {}
        for bait in possible_baits:
            distances[bait] = editdistance.eval(target_lemma, bait)
        baits = [
            bait
            for bait, distance in distances.items()
            if distance == min(list(distances.values()))
        ]
        bait = random.choice(baits)
        return [bait]

    def _get_baits(self, target_lemma: str, pos: str) -> List[str]:
        if not pos_validators.is_verb(pos):
            raise WordNotVerb()
        possible_baits = self.content_repository.get_lexemas_from_lemma(target_lemma)
        if not possible_baits:
            raise NoLexemasFound()
        bait = self._choose_bait(possible_baits, target_lemma)
        return bait

    def _get_phrase(self, original_phrase: str, target_word: str) -> str:
        return self._remove_word(original_phrase, target_word)



from .future_continuous.future_continuous_choice_question_creator import (
    FutureContinuousChoiceQuestionCreator,
)
from .future_continuous.future_continuous_negative_choice_question_creator import (
    FutureContinuousNegativeChoiceQuestionCreator,
)
from .future_perfect.future_perfect_choice_question_creator import (
    FuturePerfectChoiceQuestionCreator,
)
from .future_perfect.future_perfect_negative_choice_question_creator import (
    FuturePerfectNegativeChoiceQuestionCreator,
)
from .future_perfect_continuous.future_perfect_continuous_choice_question_creator import (
    FuturePerfectContinuousChoiceQuestionCreator,
)
from .future_perfect_continuous.future_perfect_continuous_negative_choice_question_creator import (
    FuturePerfectContinuousNegativeChoiceQuestionCreator,
)
from .going_to_future.future_dual_video_question_creator import (
    FutureDualVideoQuestionCreator,
)
from .going_to_future.going_to_future_choice_question_creator import (
    GoingToFutureChoiceQuestionCreator,
)
from .going_to_future.going_to_future_negative_choice_question_creator import (
    GoingToFutureNegativeChoiceQuestionCreator,
)
from .irregular_past_tense.irregular_past_tense_choice_question_creator import (
    IrregularPastTenseChoiceQuestionCreator,
)
from .passive_voice.passive_voice_choice_question_creator import (
    PassiveVoiceChoiceQuestionCreator,
)
from .past_continuous.past_continuous_choice_question_creator import (
    PastContinuousChoiceQuestionCreator,
)
from .past_continuous.past_continuous_negation_choice_question_creator import (
    PastContinuousNegationChoiceQuestionCreator,
)
from .past_perfect.past_perfect_choice_question_creator import (
    PastPerfectChoiceQuestionCreator,
)
from .past_perfect.past_perfect_negation_choice_question_creator import (
    PastPerfectNegationChoiceQuestionCreator,
)
from .past_perfect_continuous.past_perfect_continuous_choice_question_creator import (
    PastPerfectContinuousChoiceQuestionCreator,
)
from .past_perfect_continuous.past_perfect_continuous_negative_choice_question_creator import (
    PastPerfectContinuousNegativeChoiceQuestionCreator,
)
from .present_continuous.present_continuous_choice_question_creator import (
    PresentContinuousChoiceQuestionCreator,
)
from .present_continuous.present_continuous_full_typing_question_creator import (
    PresentContinuousFullWordTypingQuestionCreator,
)
from .present_continuous.present_continuous_interrogation_partword_typing_question_creator import (
    PresentContinuousInterrogationPartWordTypingQuestionCreator,
)
from .present_continuous.present_continuous_negative_choice_question_creator import (
    PresentContinuousNegativeChoiceQuestionCreator,
)
from .present_continuous.present_continuous_negative_full_typing_question_creator import (
    PresentContinuousNegativeFullWordTypingQuestionCreator,
)
from .present_continuous.present_continuous_negative_partword_typing_question_creator import (
    PresentContinuousNegativePartWordTypingQuestionCreator,
)
from .present_continuous.present_continuous_partword_typing_question_creator import (
    PresentContinuousPartWordTypingQuestionCreator,
)
from .present_perfect.present_perfect_choice_question_creator import (
    PresentPerfectChoiceQuestionCreator,
)
from .present_perfect.present_perfect_negation_choice_question_creator import (
    PresentPerfectNegationChoiceQuestionCreator,
)
from .present_perfect_continuous.present_perfect_continuous_choice_question_creator import (
    PresentPerfectContinuousChoiceQuestionCreator,
)
from .present_perfect_continuous.present_perfect_continuous_negative_choice_question_creator import (
    PresentPerfectContinuousNegativeChoiceQuestionCreator,
)
from .simple_future.simple_future_choice_question_creator import (
    SimpleFutureChoiceQuestionCreator,
)
from .simple_future.simple_future_choice_two_question_creator import (
    SimpleFutureChoiceTwoQuestionCreator,
)
from .simple_future.simple_future_negation_choice_question_creator import (
    SimpleFutureNegationChoiceQuestionCreator,
)
from .simple_past.simple_past_negation_choice_question_creator import (
    SimplePastNegationChoiceQuestionCreator,
)
from .simple_present.simple_present_negation_choice_question_creator import (
    SimplePresentNegationChoiceQuestionCreator,
)
from .simple_present.simple_tense_choice_question_creator import (
    SimpleTenseChoiceQuestionCreator,
)



import random

from ...dictionary_factory import pos2words
from ...dictionary_factory import word2soundslike
from ...exceptions.dictionary_exceptions import WordNotInSoundsLikeDictionary
from ...questions.question import N_BAITS
from ...questions.question import Question
from ..question_creator import QuestionCreator
from question_builder.data import DataQuestion
from typing import List


class GapQuestionCreator(QuestionCreator):

    code = "GAP"
    baits_sounds_code = "sounds"
    baits_pos_code = "pos"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:

        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word
        pos = data_question.pos

        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = target_word
        baits_code, question.baits = self._get_baits(target_word, pos)
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, target_word)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = baits_code
        return question

    def _get_baits(self, target_word, pos) -> List[str]:
        if target_word in word2soundslike:
            return (
                self.baits_sounds_code,
                random.sample(word2soundslike[target_word], N_BAITS),
            )
        else:
            return self.baits_pos_code, random.sample(pos2words[pos], N_BAITS)

    def _get_phrase(self, original_phrase, target_word) -> str:
        return self._remove_word(original_phrase, target_word)



import enum

CONTENT_KEY = "content"
CONTENT_BAIT_KEY = "bait_content"
CONTENT_PAIR_KEY = "pair_content"
INTEREST_KEY = "interest"
WORD_KEY = "word"
LEMMA_KEY = "lemma"
PAIR_KEY = "pair"
SOURCE_KEY = "source"
GETYARN_KEY = "getyarn"
TWITTER_KEY = "twitter"
ID_KEY = "id"
POS_KEY = "pos"
TAG_KEY = "tag"
PHRASE_KEY = "phrase"
TRANSLATION_KEY = "translation"
IMAGE_LINK_KEY = "image_link"
MEDIA_TYPE_KEY = "media_type"
TWEET_URL_KEY = "tweet_url"
INTEREST_LABEL_KEY = "interest_label"

TEXT = "text"
N_INTEREST_TYPES = 3
MASTERED_KEY = "mastered"
LEVEL_KEY = "level"
CONTENT_LIST_KEY = "content_list"
VERBGAMES_ITEMS_KEY = "verbgames_pattern_items"
LEMMA_CONJUGATIONS_KEY = "lemma_conjugations"
DEFAULT_N_QUESTIONS_PER_INTEREST_TYPE = 3
DEFAULT_N_QUESTIONS = 1

Q_VARIABLE_KEY = "qu"


class QuestionMode(enum.Enum):
    N_QUESTIONS_MODE = 1
    N_QUESTIONS_PER_WORD_MODE = 2



import random

from ...dictionary_factory import word2spanishdefinitions
from ...exceptions.dual_video_exceptions import ContentNotFoundException
from ...exceptions.dual_video_exceptions import DefinitionNotFoundException
from ...exceptions.dual_video_exceptions import SpanishBaitsNotFoundException
from ...questions.dual_video_question import DualVideoQuestion
from ...questions.question import N_BAITS
from ...questions.question import Question
from ...validators.pos_validators import ADJ
from ...validators.pos_validators import ADV
from ...validators.pos_validators import NOUN
from ...validators.pos_validators import VERB
from ..dual_video_question_creator import DualVideoQuestionCreator
from ..dual_video_question_creator import PROBABILITY_OF_SHUFFLING_OPTIONS_IN_DUAL_VIDEO
from question_builder.data import DataQuestion


class DualVideoSpanishDefinitionsQuestionCreator(DualVideoQuestionCreator):

    code = "DVS"
    baits_code = "nobaits"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:

        target_lemma = data_question.target_lemma
        target_word = data_question.target_word
        incorrect_content = data_question.bait_content
        correct_content = data_question.content
        definition = self._get_definition(target_lemma)

        question = DualVideoQuestion()
        question.content_id = correct_content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        medias_list = self._get_medias([correct_content, incorrect_content])
        question.links = medias_list.links
        question.media_types = medias_list.media_types
        question.phrases = medias_list.phrases
        question.correct_values = medias_list.correct_values
        question.definition = definition
        question.options = None
        question.phrase = definition
        question.original_phrase = correct_content.phrase
        question.phrase_translation = ""
        question.question_type = self.code
        question.baits_type = self.baits_code
        return question

    def _get_definition(self, target_lemma) -> str:

        if (target_lemma, NOUN) in word2spanishdefinitions:
            key = (target_lemma, NOUN)
        elif (target_lemma, VERB) in word2spanishdefinitions:
            key = (target_lemma, VERB)
        elif (target_lemma, ADV) in word2spanishdefinitions:
            key = (target_lemma, ADV)
        elif (target_lemma, ADJ) in word2spanishdefinitions:
            key = (target_lemma, ADJ)
        else:
            raise DefinitionNotFoundException()

        return word2spanishdefinitions[key][0]





from collections import defaultdict

from .. import config
from ..config import QuestionMode
from ..domain.data_converters import to_verb_question
from ..domain.lemma_data import LemmaData
from neo4j import DirectDriver
from neo4j import BoltStatementResult
from typing import List, Dict


class QuestionGetter:
    """
    Handles all the db operations related to user content management
    """

    def __init__(self, driver: DirectDriver, mode: str):
        """
        Initializes the neo4j db
        """
        self._driver = driver
        self.mode = mode
        self.mode_function_registry = {
            QuestionMode.N_QUESTIONS_MODE: self._get_n_questions,
            QuestionMode.N_QUESTIONS_PER_WORD_MODE: self._get_n_questions_per_word,
        }

    def get(self, user_id: str, lemma_list: List[str], n_questions: int, n_questions_per_word: int) -> Dict[str, List[LemmaData]]:
        with self._driver.session() as session:

            lemma_information = self._get_lemma_information(
                session, user_id, lemma_list
            )

            lemma_to_questions = defaultdict(list)
            for res in session.read_transaction(
                self.mode_function_registry[self.mode],
                user_id,
                list(lemma_information.keys()),
                n_questions,
                n_questions_per_word,
            ):
                content_rel_json = res[config.CONTENT_KEY]
                target_lemma = content_rel_json[config.LEMMA_KEY]
                content_rel_json[config.VERBGAMES_ITEMS_KEY][
                    config.LEMMA_CONJUGATIONS_KEY
                ] = lemma_information[target_lemma][config.LEMMA_CONJUGATIONS_KEY]
                verb_question = to_verb_question(content_rel_json)
                level = lemma_information[target_lemma][config.LEVEL_KEY]
                mastered = lemma_information[target_lemma][config.MASTERED_KEY]
                lemma_data = LemmaData(verb_question, level, mastered)
                lemma_to_questions[target_lemma].append(lemma_data)
            return lemma_to_questions

    def _get_lemma_information(self, session, user_id: str, lemma_list: List[str]) -> Dict:

        lemma_dict = {}
        for res in session.read_transaction(self._get_lemma_data, user_id, lemma_list):

            target_lemma = res[config.LEMMA_KEY]
            mastered = res[config.MASTERED_KEY]
            level = res[config.LEVEL_KEY]
            lemma_conjugations = res[config.LEMMA_CONJUGATIONS_KEY]

            lemma_dict[target_lemma] = {
                config.LEVEL_KEY: level,
                config.MASTERED_KEY: mastered,
                config.LEMMA_CONJUGATIONS_KEY: lemma_conjugations,
            }

        return lemma_dict

    @classmethod
    def _get_n_questions(
        cls, tx, user_id: str, lemma_list: List[str], n_questions: int, n_questions_per_word: int
    ) -> BoltStatementResult:

        return tx.run(
            cls.query_n_data_questions,
            user_id=user_id,
            lemma_list=lemma_list,
            n_questions=n_questions,
        )

    @classmethod
    def _get_n_questions_per_word(
        cls, tx, user_id: str, lemma_list: List[str], n_questions: int, n_questions_per_word: int
    ) -> BoltStatementResult:
        return tx.run(
            cls.query_n_data_questions_per_word,
            user_id=user_id,
            lemma_list=lemma_list,
            n_questions_per_word=n_questions_per_word,
        )

    @classmethod
    def _get_lemma_data(
        cls, tx, user_id: str, lemma_list: List
    ) -> BoltStatementResult:
        return tx.run(cls.query_lemma_data, user_id=user_id, lemma_list=lemma_list)



from typing import Dict, List, Optional
from .question_creators.question_creator import QuestionCreator

from .question_creators.general.antonym_question_creator import AntonymQuestionCreator
from .question_creators.general.choice_english_definitions_question_creator import (
    ChoiceEnglishDefinitionsQuestionCreator,
)
from .question_creators.general.choice_spanish_definitions_question_creator import (
    ChoiceSpanishDefinitionsQuestionCreator,
)
from .question_creators.general.conjugation_question_creator import (
    ConjugationQuestionCreator,
)
from .question_creators.general.dual_video_spanish_definitions_question_creator import (
    DualVideoSpanishDefinitionsQuestionCreator,
)
from .question_creators.general.fullword_scrabble_question_creator import (
    FullWordScrabbleQuestionCreator,
)
from .question_creators.general.fullword_typing_question_creator import (
    FullWordTypingQuestionCreator,
)
from .question_creators.general.gap_question_creator import GapQuestionCreator
from .question_creators.general.intruder_question_creator import IntruderQuestionCreator
from .question_creators.general.means_like_question_creator import (
    MeansLikeQuestionCreator,
)
from .question_creators.general.minimal_pairs_question_creator import (
    MinimalPairsQuestionCreator,
)
from .question_creators.general.multisynonym_question_creator import (
    MultiSynonymQuestionCreator,
)
from .question_creators.general.partword_typing_question_creator import (
    PartWordTypingQuestionCreator,
)
from .question_creators.general.relnoun_question_creator import RelNounQuestionCreator
from .question_creators.general.sentence_scrabble_question_creator import (
    SentenceScrabbleQuestionCreator,
)
from .question_creators.general.subword_question_creator import SubwordQuestionCreator
from .question_creators.general.synonym_question_creator import SynonymQuestionCreator
from .question_creators.general.tense_verb_choice_question_creator import (
    TenseVerbChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.future_continuous.future_continuous_choice_question_creator import (
    FutureContinuousChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.future_continuous.future_continuous_negative_choice_question_creator import (
    FutureContinuousNegativeChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.future_perfect.future_perfect_choice_question_creator import (
    FuturePerfectChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.future_perfect.future_perfect_negative_choice_question_creator import (
    FuturePerfectNegativeChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.future_perfect_continuous.future_perfect_continuous_choice_question_creator import (
    FuturePerfectContinuousChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.future_perfect_continuous.future_perfect_continuous_negative_choice_question_creator import (
    FuturePerfectContinuousNegativeChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.going_to_future.future_dual_video_question_creator import (
    FutureDualVideoQuestionCreator,
)
from .question_creators.patterns.verbs.going_to_future.going_to_future_choice_question_creator import (
    GoingToFutureChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.going_to_future.going_to_future_negative_choice_question_creator import (
    GoingToFutureNegativeChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.irregular_past_tense.irregular_past_tense_choice_question_creator import (
    IrregularPastTenseChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.passive_voice.passive_voice_choice_question_creator import (
    PassiveVoiceChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.past_continuous.past_continuous_choice_question_creator import (
    PastContinuousChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.past_continuous.past_continuous_negation_choice_question_creator import (
    PastContinuousNegationChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.past_perfect.past_perfect_choice_question_creator import (
    PastPerfectChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.past_perfect.past_perfect_negation_choice_question_creator import (
    PastPerfectNegationChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.past_perfect_continuous.past_perfect_continuous_choice_question_creator import (
    PastPerfectContinuousChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.past_perfect_continuous.past_perfect_continuous_negative_choice_question_creator import (
    PastPerfectContinuousNegativeChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.present_continuous.present_continuous_choice_question_creator import (
    PresentContinuousChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.present_continuous.present_continuous_full_typing_question_creator import (
    PresentContinuousFullWordTypingQuestionCreator,
)
from .question_creators.patterns.verbs.present_continuous.present_continuous_interrogation_partword_typing_question_creator import (
    PresentContinuousInterrogationPartWordTypingQuestionCreator,
)
from .question_creators.patterns.verbs.present_continuous.present_continuous_negative_choice_question_creator import (
    PresentContinuousNegativeChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.present_continuous.present_continuous_negative_full_typing_question_creator import (
    PresentContinuousNegativeFullWordTypingQuestionCreator,
)
from .question_creators.patterns.verbs.present_continuous.present_continuous_negative_partword_typing_question_creator import (
    PresentContinuousNegativePartWordTypingQuestionCreator,
)
from .question_creators.patterns.verbs.present_continuous.present_continuous_partword_typing_question_creator import (
    PresentContinuousPartWordTypingQuestionCreator,
)
from .question_creators.patterns.verbs.present_perfect.present_perfect_choice_question_creator import (
    PresentPerfectChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.present_perfect.present_perfect_negation_choice_question_creator import (
    PresentPerfectNegationChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.present_perfect_continuous.present_perfect_continuous_choice_question_creator import (
    PresentPerfectContinuousChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.present_perfect_continuous.present_perfect_continuous_negative_choice_question_creator import (
    PresentPerfectContinuousNegativeChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.simple_future.simple_future_choice_question_creator import (
    SimpleFutureChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.simple_future.simple_future_choice_two_question_creator import (
    SimpleFutureChoiceTwoQuestionCreator,
)
from .question_creators.patterns.verbs.simple_future.simple_future_negation_choice_question_creator import (
    SimpleFutureNegationChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.simple_past.simple_past_negation_choice_question_creator import (
    SimplePastNegationChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.simple_present.simple_present_negation_choice_question_creator import (
    SimplePresentNegationChoiceQuestionCreator,
)
from .question_creators.patterns.verbs.simple_present.simple_tense_choice_question_creator import (
    SimpleTenseChoiceQuestionCreator,
)

QUESTION_PACK_TEST_CODE = "GMPT"
QUESTION_PACK_ALL_CODE = "GMPALL"

question_type_to_creator = {
    GapQuestionCreator.code: GapQuestionCreator,
    MinimalPairsQuestionCreator.code: MinimalPairsQuestionCreator,
    SubwordQuestionCreator.code: SubwordQuestionCreator,
    PartWordTypingQuestionCreator.code: PartWordTypingQuestionCreator,
    DualVideoSpanishDefinitionsQuestionCreator.code: DualVideoSpanishDefinitionsQuestionCreator,
    ChoiceSpanishDefinitionsQuestionCreator.code: ChoiceSpanishDefinitionsQuestionCreator,
    SynonymQuestionCreator.code: SynonymQuestionCreator,
    AntonymQuestionCreator.code: AntonymQuestionCreator,
    MeansLikeQuestionCreator.code: MeansLikeQuestionCreator,
    RelNounQuestionCreator.code: RelNounQuestionCreator,
    MultiSynonymQuestionCreator.code: MultiSynonymQuestionCreator,
    IntruderQuestionCreator.code: IntruderQuestionCreator,
    ChoiceEnglishDefinitionsQuestionCreator.code: ChoiceEnglishDefinitionsQuestionCreator,
    FullWordScrabbleQuestionCreator.code: FullWordScrabbleQuestionCreator,
    SentenceScrabbleQuestionCreator.code: SentenceScrabbleQuestionCreator,
    FullWordTypingQuestionCreator.code: FullWordTypingQuestionCreator,
    ConjugationQuestionCreator.code: ConjugationQuestionCreator,
    TenseVerbChoiceQuestionCreator.code: TenseVerbChoiceQuestionCreator,
    SimpleTenseChoiceQuestionCreator.code: SimpleTenseChoiceQuestionCreator,
    GoingToFutureChoiceQuestionCreator.CODE: GoingToFutureChoiceQuestionCreator,
    SimplePresentNegationChoiceQuestionCreator.CODE: SimplePresentNegationChoiceQuestionCreator,
    SimplePastNegationChoiceQuestionCreator.CODE: SimplePastNegationChoiceQuestionCreator,
    IrregularPastTenseChoiceQuestionCreator.CODE: IrregularPastTenseChoiceQuestionCreator,
    PastContinuousChoiceQuestionCreator.CODE: PastContinuousChoiceQuestionCreator,
    PastContinuousNegationChoiceQuestionCreator.CODE: PastContinuousNegationChoiceQuestionCreator,
    PresentContinuousNegativeChoiceQuestionCreator.CODE: PresentContinuousNegativeChoiceQuestionCreator,
    PresentPerfectChoiceQuestionCreator.CODE: PresentPerfectChoiceQuestionCreator,
    PresentPerfectNegationChoiceQuestionCreator.CODE: PresentPerfectNegationChoiceQuestionCreator,
    PresentPerfectContinuousNegativeChoiceQuestionCreator.code: PresentPerfectContinuousNegativeChoiceQuestionCreator,
    PresentPerfectContinuousChoiceQuestionCreator.code: PresentPerfectContinuousChoiceQuestionCreator,
    PassiveVoiceChoiceQuestionCreator.code: PassiveVoiceChoiceQuestionCreator,
    SimpleFutureNegationChoiceQuestionCreator.CODE: SimpleFutureNegationChoiceQuestionCreator,
    PastPerfectChoiceQuestionCreator.CODE: PastPerfectChoiceQuestionCreator,
    PastPerfectNegationChoiceQuestionCreator.CODE: PastPerfectNegationChoiceQuestionCreator,
    FuturePerfectChoiceQuestionCreator.CODE: FuturePerfectChoiceQuestionCreator,
    FuturePerfectNegativeChoiceQuestionCreator.CODE: FuturePerfectNegativeChoiceQuestionCreator,
    PastPerfectContinuousChoiceQuestionCreator.CODE: PastPerfectContinuousChoiceQuestionCreator,
    PastPerfectContinuousNegativeChoiceQuestionCreator.CODE: PastPerfectContinuousNegativeChoiceQuestionCreator,
    SimpleFutureChoiceQuestionCreator.code: SimpleFutureChoiceQuestionCreator,
    SimpleFutureChoiceTwoQuestionCreator.code: SimpleFutureChoiceTwoQuestionCreator,
    FutureContinuousChoiceQuestionCreator.code: FutureContinuousChoiceQuestionCreator,
    PresentContinuousChoiceQuestionCreator.code: PresentContinuousChoiceQuestionCreator,
    FutureContinuousNegativeChoiceQuestionCreator.code: FutureContinuousNegativeChoiceQuestionCreator,
    FutureDualVideoQuestionCreator.code: FutureDualVideoQuestionCreator,
    FuturePerfectContinuousChoiceQuestionCreator.code: FuturePerfectContinuousChoiceQuestionCreator,
    FuturePerfectContinuousNegativeChoiceQuestionCreator.code: FuturePerfectContinuousNegativeChoiceQuestionCreator,
    GoingToFutureNegativeChoiceQuestionCreator.code: GoingToFutureNegativeChoiceQuestionCreator,
    PresentContinuousPartWordTypingQuestionCreator.code: PresentContinuousPartWordTypingQuestionCreator,
    PresentContinuousNegativePartWordTypingQuestionCreator.code: PresentContinuousNegativePartWordTypingQuestionCreator,
    PresentContinuousFullWordTypingQuestionCreator.code: PresentContinuousFullWordTypingQuestionCreator,
    PresentContinuousNegativeFullWordTypingQuestionCreator.code: PresentContinuousNegativeFullWordTypingQuestionCreator,
    PresentContinuousInterrogationPartWordTypingQuestionCreator.code: PresentContinuousInterrogationPartWordTypingQuestionCreator,
}


class CreatorNode:
    def __init__(self, question_creator: QuestionCreator):
        self.question_creator = question_creator
        self.next: QuestionCreator = None


class QuestionPack:
    def __init__(self, code: str, level_to_question_types: Dict[int, List[str]], level_to_master: int = None):
        self.code: str = code
        self.level_to_question_types: Dict[int, List[str]] = level_to_question_types
        self.level_to_master: int = len(level_to_question_types.keys())
        self._question_types: List[str] = None
        self._question_type_to_level: Dict[str, int] = None
        self._responsibility_chain: Dict[str, CreatorNode] = None

    @property
    def question_types(self) -> List[str]:
        if self._question_types is None:
            self._question_types = []
            for level, question_types in self.level_to_question_types.items():
                for question_type in question_types:
                    self._question_types.append(question_type)
        return self._question_types

    @property
    def question_type_to_level(self) -> Dict[str, int]:
        if self._question_type_to_level is None:
            self._question_type_to_level = {}
            for level, question_types in self.level_to_question_types.items():
                for question_type in question_types:
                    self._question_type_to_level[question_type] = level
        return self._question_type_to_level

    @property
    def responsibility_chain(self) -> Dict[str, CreatorNode]:
        """
        Create the responsability chain according to the question pack given
        in the constructor. The chain is a single linked list.
        """
        if self._responsibility_chain is None:
            self._responsibility_chain = {}
            levels = sorted(list(self.level_to_question_types.keys()), reverse=True)
            last = None
            for level in levels:
                questions = self.level_to_question_types[level]
                for index in range(len(questions)):
                    #  Get the current node in the chain
                    question_code = questions[index]
                    question_creator = question_type_to_creator[question_code]
                    self._responsibility_chain[question_code] = CreatorNode(
                        question_creator
                    )
                    #  Get the previous node and link it to the current one
                    if last:
                        last.next = self._responsibility_chain[question_code]
                    last = self._responsibility_chain[question_code]
        return self._responsibility_chain



from question_builder.data.question_type_repository.verbs.future_continuous.FCC_question_getter import (
    FCCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.future_continuous.FCNC_question_getter import (
    FCNCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.future_perfect.FPC_question_getter import (
    FPCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.future_perfect.FPNC_question_getter import (
    FPNCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.future_perfect_continuous.FPCC_question_getter import (
    FPCCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.future_perfect_continuous.FPCNC_question_getter import (
    FPCNCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.going_to_future.FDV_question_getter import (
    FDVQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.going_to_future.GTFC_question_getter import (
    GTFCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.going_to_future.GTFNC_question_getter import (
    GTFNCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.irregular_past_tense.IRTCP_question_getter import (
    IRTCPQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.passive_voice.PVC_question_getter import (
    PVCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.past_continuous.PACC_question_getter import (
    PACCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.past_continuous.PACNC_question_getter import (
    PACNCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.past_perfect.PAPC_question_getter import (
    PAPCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.past_perfect.PAPNC_question_getter import (
    PAPNCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.past_perfect_continuous.PPCC_question_getter import (
    PPCCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.past_perfect_continuous.PPCNC_question_getter import (
    PPCNCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.present_continuous.PCC_question_getter import (
    PCCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.present_continuous.PCFT_question_getter import (
    PCFTQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.present_continuous.PCIPT_question_getter import (
    PCIPTQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.present_continuous.PCNC_question_getter import (
    PCNCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.present_continuous.PCNFT_question_getter import (
    PCNFTQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.present_continuous.PCNPT_question_getter import (
    PCNPTQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.present_continuous.PCPT_question_getter import (
    PCPTQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.present_perfect.PPC_question_getter import (
    PPCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.present_perfect.PPNC_question_getter import (
    PPNCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.present_perfect_continuous.PRPCC_question_getter import (
    PRPCCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.present_perfect_continuous.PRPCNC_question_getter import (
    PRPCNCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.simple_future.SFC2_question_getter import (
    SFC2QuestionGetter,
)
from question_builder.data.question_type_repository.verbs.simple_future.SFC_question_getter import (
    SFCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.simple_future.SFNC_question_getter import (
    SFNCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.simple_past.SPANC_question_getter import (
    SPANCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.simple_present.SPNC_question_getter import (
    SPNCQuestionGetter,
)
from question_builder.data.question_type_repository.verbs.simple_present.STC_question_getter import (
    STCQuestionGetter,
)



import pathlib
import pickle

PACKAGE_ROOT = pathlib.Path(__file__).resolve().parent.parent

POS_DICTIONARY_FILE = PACKAGE_ROOT / "resources" / "pos_dictionary.pk"
SOUND_DICTIONARY_FILE = PACKAGE_ROOT / "resources" / "sounds_dictionary.pk"
SYNONYM_DICTIONARY_FILE = PACKAGE_ROOT / "resources" / "synonym_dictionary.pk"
ANTONYM_DICTIONARY_FILE = PACKAGE_ROOT / "resources" / "antonym_dictionary.pk"
MEANSLIKE_DICTIONARY_FILE = PACKAGE_ROOT / "resources" / "meanslike_dictionary.pk"
PRESUFF_DICTIONARY_FILE = PACKAGE_ROOT / "resources" / "presuff_dictionary.pk"
INTRUDER_DICTIONARY_FILE = PACKAGE_ROOT / "resources" / "intruder_dictionary.pk"
RELNOUN_DICTIONARY_FILE = PACKAGE_ROOT / "resources" / "relnoun_dictionary.pk"
PARTWORDTYPING_DICTIONARY_FILE = (
    PACKAGE_ROOT / "resources" / "partwordtyping_dictionary.pk"
)
MINIMAL_PAIR_DICTIONARY_FILE = (
    PACKAGE_ROOT / "resources" / "minimal_pairs_dictionary.pk"
)
MULTISYNONYM_DICTIONARY_FILE = PACKAGE_ROOT / "resources" / "multisynonym_dictionary.pk"
SPANISH_DEFINITIONS_DICTIONARY_FILE = (
    PACKAGE_ROOT / "resources" / "spanish_definitions_dictionary.pk"
)
ENGLISH_DEFINITIONS_DICTIONARY_FILE = (
    PACKAGE_ROOT / "resources" / "english_definitions_dictionary.pk"
)

CORRECT = "correct"
BAITS = "baits"

with open(POS_DICTIONARY_FILE, "rb") as config_dictionary_file:
    pos2words = pickle.load(config_dictionary_file)

with open(SYNONYM_DICTIONARY_FILE, "rb") as config_dictionary_file:
    word2synonyms = pickle.load(config_dictionary_file)

with open(SOUND_DICTIONARY_FILE, "rb") as config_dictionary_file:
    word2soundslike = pickle.load(config_dictionary_file)

with open(ANTONYM_DICTIONARY_FILE, "rb") as config_dictionary_file:
    word2antonym = pickle.load(config_dictionary_file)

with open(MEANSLIKE_DICTIONARY_FILE, "rb") as config_dictionary_file:
    word2meanslike = pickle.load(config_dictionary_file)

with open(PRESUFF_DICTIONARY_FILE, "rb") as config_dictionary_file:
    word2presuffix = pickle.load(config_dictionary_file)

with open(INTRUDER_DICTIONARY_FILE, "rb") as config_dictionary_file:
    word2intruder = pickle.load(config_dictionary_file)

with open(RELNOUN_DICTIONARY_FILE, "rb") as config_dictionary_file:
    word2relnoun = pickle.load(config_dictionary_file)

with open(PARTWORDTYPING_DICTIONARY_FILE, "rb") as config_dictionary_file:
    word2partwordtyping = pickle.load(config_dictionary_file)

with open(MINIMAL_PAIR_DICTIONARY_FILE, "rb") as config_dictionary_file:
    word2minimalpair = pickle.load(config_dictionary_file)

with open(MULTISYNONYM_DICTIONARY_FILE, "rb") as config_dictionary_file:
    word2multisynonym = pickle.load(config_dictionary_file)

with open(SPANISH_DEFINITIONS_DICTIONARY_FILE, "rb") as config_dictionary_file:
    word2spanishdefinitions = pickle.load(config_dictionary_file)

with open(ENGLISH_DEFINITIONS_DICTIONARY_FILE, "rb") as config_dictionary_file:
    word2englishdefinitions = pickle.load(config_dictionary_file)



import re

from ...config import MAX_SENTENCE_SCRABBLE_LENGTH
from ...config import MIN_SENTENCE_SCRABBLE_LENGTH
from ...exceptions.dual_video_exceptions import MaximumLengthExceeded
from ...questions.question import CORRECT
from ...questions.question import Question
from ...questions.question import TEXT
from ..question_creator import QuestionCreator
from question_builder.data import DataQuestion
from typing import List, Dict


class SentenceScrabbleQuestionCreator(QuestionCreator):

    code = "SESB"
    baits_code = "nobaits"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:
        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word

        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = self._get_correct_answer(content.phrase)
        question.baits = self._get_baits()
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        return question

    def _get_correct_answer(self, original_phrase: str) -> List[str]:
        length = len(original_phrase.split())
        if MIN_SENTENCE_SCRABBLE_LENGTH < length < MAX_SENTENCE_SCRABBLE_LENGTH:
            return [token for token in original_phrase.split()[1:-1] if len(token) > 0]
        raise MaximumLengthExceeded()

    def _get_baits(self) -> List[str]:
        return []

    def _get_options(self, correct_answer: List[str], baits: List[str]) -> List[Dict]:
        """
        Get the baits and the words, and make an array of Options Objects in
        which word is set to True, and all others are False
        """
        options = []
        for answer in correct_answer:
            options.append({TEXT: answer, CORRECT: True})
        for bait in baits:
            options.append({TEXT: bait, CORRECT: False})
        return options

    def _get_phrase(self, original_phrase: str) -> str:
        return self._remove_words(original_phrase)

    def _remove_words(self, original_phrase: str) -> str:
        complete_phrase = original_phrase.split()
        inner_phrase = re.sub(
            r"[^ ]", r"_", " ".join(complete_phrase[1:-1]), flags=re.I
        )
        phrase = " ".join([complete_phrase[0], inner_phrase, complete_phrase[-1]])
        phrase = self._clean_phrase(phrase)
        return phrase



GET_QUESTION_PACKS = """
MATCH (questionPack:QUESTION_PACK) - [rel:HAS_QUESTION_TYPE] -> (questionType:QUESTION_TYPE)
WITH questionPack, rel.level as level, questionType order by level, rel.priority
RETURN questionPack as question_pack, level, collect(questionType) as question_types order by questionPack.code
"""



from ...question_getter import QuestionGetter
from .future_perfect_affirmative_queries import GET_N_DATA_QUESTIONS
from .future_perfect_affirmative_queries import GET_N_DATA_QUESTIONS_PER_WORD
from .lemma_queries import GET_LEMMA_LIST_DATA


class FPCQuestionGetter(QuestionGetter):
    code = "FPC"
    query_n_data_questions_per_word = GET_N_DATA_QUESTIONS_PER_WORD
    query_n_data_questions = GET_N_DATA_QUESTIONS
    query_lemma_data = GET_LEMMA_LIST_DATA



import random

from question_builder.bp.dictionary_factory import word2englishdefinitions
from question_builder.bp.exceptions.dictionary_exceptions import (
    WordNotInEnglishDefinitionsDictionary,
)
from question_builder.bp.question_creators.question_creator import QuestionCreator
from question_builder.bp.questions.question import Question
from question_builder.data import DataQuestion
from typing import List

SUBJECT_KEY = "subject"
CONJUGATED_AUXILIARYVERB_KEY = "conjugated_auxiliaryverb"
LEMMA_CONJUGATIONS = "lemma_conjugations"
PRESENT_Continuous_KEY = "VBG"
PAST_TENSE_KEY = "VBD"
GOING = "going"
GOING_TO = "going to"
SAMPLE_N_BAITS = 2


class GoingToFutureChoiceQuestionCreator(QuestionCreator):

    CODE = "GTFC"
    BAITS_CODE = "nogtfc"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:
        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word

        verbgames_pattern_items = data_question.verbgames_pattern_items
        conjugated_verbtobe = verbgames_pattern_items[CONJUGATED_AUXILIARYVERB_KEY]
        target_verb_present_continuous = verbgames_pattern_items[LEMMA_CONJUGATIONS][
            PRESENT_Continuous_KEY
        ]
        target_verb_past_tense = verbgames_pattern_items[LEMMA_CONJUGATIONS][
            PAST_TENSE_KEY
        ]
        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = self._get_correct_answer(
            target_word, conjugated_verbtobe
        )
        question.baits = self._get_baits(
            target_word,
            conjugated_verbtobe,
            target_verb_present_continuous,
            target_verb_past_tense,
        )
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(
            content.phrase, target_word, conjugated_verbtobe
        )
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.CODE
        question.baits_type = self.BAITS_CODE
        return question

    @staticmethod
    def _get_correct_answer(target_verb: str, conjugated_verbtobe: str) -> str:
        return f"{conjugated_verbtobe} {GOING_TO} {target_verb}"

    @staticmethod
    def _get_baits(
        target_verb: str,
        conjugated_verbtobe: str,
        target_verb_present_continuous: str,
        target_verb_past_tense: str,
    ) -> List[str]:
        return random.sample(
            [
                f"{conjugated_verbtobe} {GOING} {target_verb}",
                f"{conjugated_verbtobe} {GOING_TO} {target_verb_past_tense}",
                f"{conjugated_verbtobe} {GOING_TO} {target_verb_present_continuous}",
            ],
            SAMPLE_N_BAITS,
        )

    def _get_phrase(self, 
                original_phrase: str, 
                target_verb: str, 
                conjugated_verbtobe: str) -> str:
        return self._underline_word(
            original_phrase, f"{conjugated_verbtobe} {GOING_TO} {target_verb}"
        )



import typing as t
from dataclasses import dataclass

from .data_question import DataQuestion


@dataclass
class LemmaData:
    data_question: DataQuestion
    level: int = 1
    mastered: bool = False



from question_builder.data.config import QuestionMode
from question_builder.data.content_repository.content_repository import (
    ContentRepository,
)
from question_builder.data.domain.content import GetyarnContent
from question_builder.data.domain.content import TwitterContent
from question_builder.data.domain.content import Content
from question_builder.data.domain.data_converters import to_content
from question_builder.data.domain.data_converters import to_question
from question_builder.data.domain.data_converters import to_verb_question
from question_builder.data.domain.data_question import DataQuestion
from question_builder.data.domain.lemma_data import LemmaData
from question_builder.data.question_pack_repository.question_pack_repository import (
    QuestionPackRepository,
)
from question_builder.data.question_type_repository.question_type_repository import (
    QuestionTypeRepository,
)



import random

from question_builder.bp.question_creators.question_creator import QuestionCreator
from question_builder.bp.questions.question import Question
from question_builder.data import DataQuestion
import re
from typing import List

SUBJECT_KEY = "subject"
CONJUGATED_VERBTOBE_KEY = "conjugated_auxiliaryverb"
NEGATION_ADVERB_KEY = "negation"
PATTERN_KEY = "pattern"
APOSTROPHE = "'"
VERBTOBE_PRESENT_1S_ENDING = "m"
VERBTOBE_PRESENT_2S_ENDING = "re"
VERBTOBE_PRESENT_3S_ENDING = "s"
VERBTOBE_PAST_1S = "was"
VERBTOBE_PAST_2S = "were"
VERBTOBE_PRESENT_1S = "am"
VERBTOBE_PRESENT_2S = "are"
VERBTOBE_PRESENT_3S = "is"
AI = "ai"
DONT = "don't"
NOT = "not"
NO = "no"
SAMPLE_N_BAITS = 2
INSTRUCTIONS = 'Elige la opción de presente continuo'


class PresentContinuousNegativeChoiceQuestionCreator(QuestionCreator):

    CODE = "PCNC"
    BAITS_CODE = "nopcnc"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:
        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word

        verbgames_pattern_items = data_question.verbgames_pattern_items
        subject = verbgames_pattern_items[SUBJECT_KEY]
        conjugated_verbtobe = verbgames_pattern_items[CONJUGATED_VERBTOBE_KEY]
        negation_adverb = verbgames_pattern_items[NEGATION_ADVERB_KEY]
        pattern = verbgames_pattern_items[PATTERN_KEY]

        subject_verbtobe_negation = self._get_combined_subject_verbtobe_negation(
            subject, conjugated_verbtobe, negation_adverb
        )
        verbtobe_past_tense = self._get_past_tense_verbtobe(conjugated_verbtobe)

        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = self._get_correct_answer(
            target_word, subject_verbtobe_negation
        )
        question.baits = self._get_baits(
            target_word, target_lemma, subject, conjugated_verbtobe, verbtobe_past_tense
        )
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, pattern)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.CODE
        question.baits_type = self.BAITS_CODE
        question.instructions = INSTRUCTIONS.format(target_lemma)
        return question

    @staticmethod
    def _get_combined_subject_verbtobe_negation(
        subject: str, conjugated_verbtobe: str, negation_adverb: str
    ) -> str:
        if conjugated_verbtobe.startswith(APOSTROPHE):
            return f"{subject}{conjugated_verbtobe} {negation_adverb}"
        elif negation_adverb.startswith(APOSTROPHE) or conjugated_verbtobe == AI:
            return f"{subject} {conjugated_verbtobe}{negation_adverb}"
        else:
            return f"{subject} {conjugated_verbtobe} {negation_adverb}"

    @staticmethod
    def _get_past_tense_verbtobe(conjugated_verbtobe: str) -> str:
        if conjugated_verbtobe.endswith(
            VERBTOBE_PRESENT_1S_ENDING
        ) or conjugated_verbtobe.endswith(VERBTOBE_PRESENT_3S_ENDING):
            return VERBTOBE_PAST_1S
        else:
            return VERBTOBE_PAST_2S

    @staticmethod
    def _get_correct_answer(target_verb: str, subject_verbtobe_negation: str) -> str:
        return f"{subject_verbtobe_negation} {target_verb}"

    def _get_baits(
        self,
        target_verb: str,
        target_lemma: str,
        subject: str,
        conjugated_verbtobe: str,
        verbtobe_past_tense: str,
    ) -> List[str]:

        verbtobe_present_no_contraction = self._get_verbtobe_present_no_contraction(
            conjugated_verbtobe
        )

        return random.sample(
            [
                f"{subject} {DONT} {verbtobe_present_no_contraction} {target_verb}",
                f"{subject} {NOT} {verbtobe_present_no_contraction} {target_verb}",
                f"{subject} {verbtobe_present_no_contraction} {NO} {target_verb}",
                f"{subject} {verbtobe_past_tense} {NOT} {target_lemma}",
            ],
            SAMPLE_N_BAITS,
        )

    def _get_phrase(self, original_phrase: str, correct_answer: str) -> str:
        redata = re.compile(re.escape(correct_answer), re.IGNORECASE)
        phrase = redata.sub(len(correct_answer)*"_" , original_phrase)
        return phrase

    @staticmethod
    def _get_verbtobe_present_no_contraction(conjugated_verbtobe: str) -> str:
        if conjugated_verbtobe.endswith(VERBTOBE_PRESENT_1S_ENDING):
            return VERBTOBE_PRESENT_1S
        elif conjugated_verbtobe.endswith(VERBTOBE_PRESENT_2S_ENDING):
            return VERBTOBE_PRESENT_2S
        else:
            return VERBTOBE_PRESENT_3S



import random
import re

from question_builder.bp.exceptions.question_creator_exception import (
    QuestionCreatorException,
)
from question_builder.bp.question_creators.question_creator import QuestionCreator
from question_builder.bp.questions.question import Question
from question_builder.data import DataQuestion
from typing import List

PATTERN_KEY = "pattern"
SUBJECT_KEY = "subject"
LEMMA_CONJUGATIONS = "lemma_conjugations"
VBG = "VBG"
BEEN = "been"
N_BAITS = 2


class PresentPerfectContinuousNegativeChoiceQuestionCreator(QuestionCreator):

    code = "PRPCNC"
    baits_code = "noprpcnc"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:

        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word

        verbgames_pattern_items = data_question.verbgames_pattern_items
        pattern = verbgames_pattern_items[PATTERN_KEY]
        subject = verbgames_pattern_items[SUBJECT_KEY]
        lemma_conjugations = verbgames_pattern_items[LEMMA_CONJUGATIONS]
        verb_present_continuous = lemma_conjugations[VBG]

        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = pattern
        question.baits = self._get_baits(subject, target_lemma, verb_present_continuous)
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, pattern)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        question.content_id = content.id
        return question

    def _get_baits(self, subject: str, target_lemma: str, verb_present_continuous: str) -> List[str]:
        return random.sample(
            [
                f"{subject} don't have {BEEN} {verb_present_continuous}",
                f"{subject} not have {BEEN} {verb_present_continuous}",
                f"{subject} have no {BEEN} {verb_present_continuous}",
            ],
            N_BAITS,
        )

    def _get_phrase(self, original_phrase: str, pattern: str) -> str:
        redata = re.compile(re.escape(pattern), re.IGNORECASE)
        phrase = redata.sub("_" * len(pattern), original_phrase)
        return phrase



from . import question
from typing import Dict

class DualVideoQuestion(question.Question):
    def to_json(self) -> Dict:
        question_ = {
            question.ID: self.id,
            question.MEDIA: [
                {
                    question.MEDIA_TYPE: media_type,
                    question.MEDIA_LINK: link,
                    question.PHRASE: phrase,
                    question.CORRECT: correct_value,
                }
                for media_type, link, phrase, correct_value in zip(
                    self.media_types, self.links, self.phrases, self.correct_values
                )
            ],
            question.PHRASE: self.phrase,
            question.ORIGINAL_PHRASE: self.original_phrase,
            question.TRANSLATION_KEY: self.phrase_translation,
            question.OPTIONS: self.options,
            question.QUESTION_TYPE: self.question_type,
            question.LEVEL: self.level,
            question.IS_LAST_LEVEL: self.is_last_level,
            question.TARGET_LEMMA: self.target_lemma,
            question.MASTERED: self.mastered,
        }
        return question_



from .. import config
from .content import GetyarnContent
from .content import TwitterContent
from .content import Content
from .data_question import DataQuestion
from typing import Dict


def to_content(content_json: Dict[str, Dict]) -> Content:
    if(content_json == None):
        return
    source = content_json[config.SOURCE_KEY]
    if source == config.GETYARN_KEY:
        content = GetyarnContent(
            id = content_json[config.ID_KEY],
            phrase = content_json[config.PHRASE_KEY],
            translation = content_json[config.TRANSLATION_KEY],
            media_type = content_json[config.MEDIA_TYPE_KEY],
            source = content_json[config.SOURCE_KEY],
            interest_label = content_json[config.INTEREST_LABEL_KEY],
        )
    elif source == config.TWITTER_KEY:
        content = TwitterContent(
            id = content_json[config.ID_KEY],
            image_link = content_json[config.IMAGE_LINK_KEY],
            phrase = content_json[config.PHRASE_KEY],
            media_type = content_json[config.MEDIA_TYPE_KEY],
            tweet_url = content_json[config.TWEET_URL_KEY],
            source = content_json[config.SOURCE_KEY],
            interest_label = content_json[config.INTEREST_LABEL_KEY],
        )
    return content


def to_question(content_rel_json: Dict) -> DataQuestion:
    content = to_content(content_rel_json[config.CONTENT_KEY])
    bait_content = to_content(content_rel_json[config.CONTENT_BAIT_KEY])
    pair_content = to_content(content_rel_json[config.CONTENT_PAIR_KEY])
    interest = content_rel_json[config.INTEREST_KEY]
    target_lemma = content_rel_json[config.LEMMA_KEY]
    target_word = content_rel_json[config.WORD_KEY]
    pos = content_rel_json[config.POS_KEY]
    return DataQuestion(content, 
                        interest, 
                        target_lemma, 
                        target_word, 
                        pos, 
                        bait_content = bait_content,
                        pair_content = pair_content)


def to_verb_question(content_rel_json: Dict) -> DataQuestion:
    content = to_content(content_rel_json[config.CONTENT_KEY])
    interest = content_rel_json[config.INTEREST_KEY]
    target_lemma = content_rel_json[config.LEMMA_KEY]
    target_word = content_rel_json[config.WORD_KEY]
    pos = content_rel_json[config.POS_KEY]
    verbgames_pattern_items = None
    if config.VERBGAMES_ITEMS_KEY in content_rel_json:
        verbgames_pattern_items = content_rel_json[config.VERBGAMES_ITEMS_KEY]
    return DataQuestion(
        content, interest, target_lemma, target_word, pos, None, verbgames_pattern_items
    )



import random
import re

from ..questions import question
from question_builder.data import DataQuestion
from typing import List, Dict, Tuple
from question_builder.data import ContentRepository, Content


class QuestionCreator:
    def __init__(self, content_repository: ContentRepository):
        self.content_repository = content_repository

    def _get_links_and_media_types(self, content: Content) -> Tuple[List[str], List[str]]:
        """
        Gets links and media types lists depending on type
        """
        if content.source == question.GETYARN_KEY:
            link_jpg = f"https://y.yarn.co/{content.id}_screenshot.jpg"
            link_mp4 = f"http://media.wordbox.ai/{content.interest_label.lower()}/{content.id}"
            links = [link_jpg, link_mp4]
            media_types = [question.IMAGE, question.VIDEO]
        if content.source == question.TWITTER_KEY:
            link_jpg = content.image_link
            links = [link_jpg]
            media_types = [question.IMAGE]
        return links, media_types

    def _get_translation(self, content: Content) -> str:
        if hasattr(content, "translation"):
            if content.translation is not None:
                phrase_translation = content.translation.strip()
            else:
                phrase_translation = ""
        else:
            phrase_translation = ""
        return phrase_translation

    def _remove_word(self, phrase: str, word: str) -> str:
        blank_line = "_" * len(word)
        phrase = re.sub(
            r"\b" + word + r"(\b|(?=n't))", r"" + blank_line + "", phrase, flags=re.I
        )
        phrase = self._clean_phrase(phrase)
        return phrase

    def _underline_word(self, phrase: str, word: str) -> str:
        splitted = re.split(word, phrase, flags=re.IGNORECASE)
        formated_word = "\u0332".join(word + " ")
        phrase = formated_word.join(splitted)
        phrase = self._clean_phrase(phrase)
        return phrase

    def _clean_phrase(self, phrase: str) -> str:
        phrase = re.sub(r"(&.*?;)", "", phrase)  # &amp; and others
        phrase = re.sub(r"(http[s]?://(.*?)( |,|;|\Z))", "... ", phrase)  # URL
        phrase = phrase.replace("\n", " ")  # \n
        return phrase

    def _get_options(self, correct_answer: str, baits: List[str]) -> List[Dict]:
        """
        Get the baits and the words, and make an array of Options Objects in
        which word is set to True, and all others are False
        """
        options = [{question.TEXT: correct_answer, question.CORRECT: True}]
        for bait in baits:
            options.append({question.TEXT: bait, question.CORRECT: False})
        random.shuffle(options)
        return options



VERB = "VERB"
AUX = "AUX"
NOUN = "NOUN"
PRON = "PRON"
PROPN = "PROPN"
ADJ = "ADJ"
ADV = "ADV"


def is_verb(pos: str) -> bool:
    return pos == VERB


def is_auxiliar(pos: str) -> bool:
    return pos == AUX


def is_noun(pos: str) -> bool:
    return pos == NOUN or pos == PROPN or pos == PRON


def is_adjective(pos: str) -> bool:
    return pos == ADJ


def is_adverb(pos: str) -> bool:
    return pos == ADV



GET_LEMMA_LIST_DATA = """
MATCH (lemma:LEMMA)
MATCH (user:USER{id:$user_id})
WHERE lemma.text in $lemma_list
OPTIONAL MATCH (lemma) -[:HAS_LEMMA{tag:"VBD"}]-(verb_vbd:WORD)
WITH user, lemma, {VBD: coalesce(collect(verb_vbd.text)[0], 'N/A')} as lemma_conjugations
OPTIONAL MATCH (user) -[has_seen:HAS_SEEN] -(lemma)
RETURN lemma.text as lemma, coalesce(has_seen.level, 1) as level, coalesce(has_seen.mastered, False) as mastered, lemma_conjugations
"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['question_builder',
 'question_builder.bp',
 'question_builder.bp.exceptions',
 'question_builder.bp.question_creators',
 'question_builder.bp.question_creators.general',
 'question_builder.bp.question_creators.patterns',
 'question_builder.bp.question_creators.patterns.verbs',
 'question_builder.bp.question_creators.patterns.verbs.future_continuous',
 'question_builder.bp.question_creators.patterns.verbs.future_perfect',
 'question_builder.bp.question_creators.patterns.verbs.future_perfect_continuous',
 'question_builder.bp.question_creators.patterns.verbs.going_to_future',
 'question_builder.bp.question_creators.patterns.verbs.irregular_past_tense',
 'question_builder.bp.question_creators.patterns.verbs.passive_voice',
 'question_builder.bp.question_creators.patterns.verbs.past_continuous',
 'question_builder.bp.question_creators.patterns.verbs.past_perfect',
 'question_builder.bp.question_creators.patterns.verbs.past_perfect_continuous',
 'question_builder.bp.question_creators.patterns.verbs.present_continuous',
 'question_builder.bp.question_creators.patterns.verbs.present_perfect',
 'question_builder.bp.question_creators.patterns.verbs.present_perfect_continuous',
 'question_builder.bp.question_creators.patterns.verbs.simple_future',
 'question_builder.bp.question_creators.patterns.verbs.simple_past',
 'question_builder.bp.question_creators.patterns.verbs.simple_present',
 'question_builder.bp.questions',
 'question_builder.bp.validators',
 'question_builder.config',
 'question_builder.data',
 'question_builder.data.content_repository',
 'question_builder.data.domain',
 'question_builder.data.question_pack_repository',
 'question_builder.data.question_type_repository',
 'question_builder.data.question_type_repository.verbs',
 'question_builder.data.question_type_repository.verbs.future_continuous',
 'question_builder.data.question_type_repository.verbs.future_perfect',
 'question_builder.data.question_type_repository.verbs.future_perfect_continuous',
 'question_builder.data.question_type_repository.verbs.going_to_future',
 'question_builder.data.question_type_repository.verbs.irregular_past_tense',
 'question_builder.data.question_type_repository.verbs.passive_voice',
 'question_builder.data.question_type_repository.verbs.past_continuous',
 'question_builder.data.question_type_repository.verbs.past_perfect',
 'question_builder.data.question_type_repository.verbs.past_perfect_continuous',
 'question_builder.data.question_type_repository.verbs.present_continuous',
 'question_builder.data.question_type_repository.verbs.present_perfect',
 'question_builder.data.question_type_repository.verbs.present_perfect_continuous',
 'question_builder.data.question_type_repository.verbs.simple_future',
 'question_builder.data.question_type_repository.verbs.simple_past',
 'question_builder.data.question_type_repository.verbs.simple_present',
 'question_builder.resources']

package_data = \
{'': ['*']}

install_requires = \
['editdistance>=0.5.3,<0.6.0',
 'neo4j==1.7.6',
 'setuptools>=50.3.0,<51.0.0',
 'wheel>=0.35.1,<0.36.0']

setup_kwargs = {
    'name': 'backend-library-createquestions',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Cesar Juarez',
    'author_email': 'cesar@wordbox.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from .content import Content
from typing import Optional, Dict

class DataQuestion:
    def __init__(
        self,
        content: Content,
        interest: str,
        target_lemma: str,
        target_word: str,
        pos: str,
        tag: Optional[str] = None,
        verbgames_pattern_items: Optional[Dict[str, str]] = None,
        bait_content: Optional[Content] = None, 
        pair_content: Optional[Content] = None,
    ):
        self.content = content
        self.interest = interest
        self.target_lemma = target_lemma
        self.target_word = target_word
        self.pos = pos
        self.tag = tag
        self.verbgames_pattern_items = verbgames_pattern_items
        self.bait_content = bait_content
        self.pair_content = pair_content

    def __repr__(self):
        return "DataQuestion({}, {} , {} , {} , {} , {} , {} )".format(
            self.content,
            self.target_lemma,
            self.target_word,
            self.pos,
            self.tag,
            self.verbgames_pattern_items,
            self.bait_content.id if self.bait_content else None,
        )



GET_CONTENT_BAITS = """
MATCH (interest) - [:HAS_CONTENT] -> (content:CONTENT) - [pattern_rel:HAS_PATTERN{lemma:$lemma}] -> (:PATTERN{code:'SFC'})
MATCH (content)-[rel_target_word:HAS_WORD{position : pattern_rel.position}]-(target_word:WORD)
WHERE (:USER{id:$user_id}) - [:LIKES] ->(interest)
WITH {
    lemma: pattern_rel.lemma,
    content: {
        id: content.id,
        translation: content.translation,
        source: content.source,
        media_type: content.media_type,
        phrase: content.phrase,
        interest_label: labels(interest)[0]
    },
    interest:interest.name,
    word:target_word.text,
    pos:rel_target_word.pos,
    verbgames_pattern_items:{
                            subject: null,
                            pattern: pattern_rel.pattern,
                            conjugated_auxiliaryverb: null
                            }
} AS content, RAND() AS rnd ORDER BY rnd LIMIT 10

RETURN content LIMIT 1
"""



import random

from ...dictionary_factory import word2spanishdefinitions
from ...exceptions.dictionary_exceptions import WordNotInSpanishDefinitionsDictionary
from ...questions.question import Question
from ..question_creator import QuestionCreator
from question_builder.data import DataQuestion
from typing import List


class ChoiceSpanishDefinitionsQuestionCreator(QuestionCreator):

    code = "CSD"
    baits_code = "nobaits"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:

        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word
        pos = data_question.pos

        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = self._get_correct_answer(target_lemma, pos)
        question.baits = self._get_baits(target_lemma)
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, target_word)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        return question

    def _get_correct_answer(self, target_lemma: str, pos: str) -> str:
        if (target_lemma, pos) in word2spanishdefinitions:
            return random.choice(word2spanishdefinitions[(target_lemma, pos)])
        raise WordNotInSpanishDefinitionsDictionary()

    def _get_baits(self, target_lemma: str) -> List[str]:
        bait = None
        while True:
            bait = random.choice(list(word2spanishdefinitions.keys()))
            if bait != target_lemma:
                break
        return [random.choice(word2spanishdefinitions[bait])]

    def _get_phrase(self, original_phrase: str, target_word: str) -> str:
        return self._underline_word(original_phrase, target_word)



import random
import re

from question_builder.bp.question_creators.question_creator import QuestionCreator
from question_builder.bp.questions.question import Question
from question_builder.data import DataQuestion
from typing import List

LEMMA_CONJUGATIONS = "lemma_conjugations"
PATTERN_KEY = "pattern"
VBG = "VBG"
AM = "am"
CONTRACTED_AM = "'m"
IS = "is"
CONTRACTED_IS = "'s"
ARE = "are"
CONTRACTED_ARE = "'re"
AINT = "ain't"
INSTRUCTIONS = 'Completa la negación'


class PresentContinuousInterrogationPartWordTypingQuestionCreator(QuestionCreator):

    code = "PCIPT"
    baits_code = "nopcipt"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:
        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word
        verbgames_pattern_items = data_question.verbgames_pattern_items
        pattern = verbgames_pattern_items[PATTERN_KEY]
        lemma_conjugations = verbgames_pattern_items[LEMMA_CONJUGATIONS]
        verb_present_continuous = lemma_conjugations[VBG]
        question = Question()
        question.content_id = content.id
        question.target_lemma = target_lemma
        question.target_word = target_word
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = self._get_correct_answer(
            verb_present_continuous, pattern
        )
        question.baits = self._get_baits()
        question.options = self._get_options(question.correct_answer, question.baits)

        question.phrase = self._get_phrase(content.phrase, question.correct_answer)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.code
        question.baits_type = self.baits_code
        question.instructions = INSTRUCTIONS.format(target_lemma)
        return question

    def _get_correct_answer(self, verb_present_continuous: str, pattern: str) -> str:
        if bool(random.getrandbits(1)):
            return verb_present_continuous
        if AM in pattern.lower():
            return AM
        if IS in pattern.lower():
            return IS
        if ARE in pattern.lower():
            return ARE
        if CONTRACTED_AM in pattern.lower():
            return CONTRACTED_AM
        if CONTRACTED_IS in pattern.lower():
            return CONTRACTED_IS
        if CONTRACTED_ARE in pattern.lower():
            return CONTRACTED_ARE
        if AINT in pattern.lower():
            return AINT

    def _get_baits(self) -> List[str]:
        return []

    def _get_phrase(self, original_phrase: str, correct_answer: str) -> str:
        return self._remove_word(original_phrase, correct_answer)



from collections import defaultdict

from ...question_getter import QuestionGetter
from .going_to_future_affirmative_queries import GET_N_DATA_QUESTIONS
from .going_to_future_affirmative_queries import GET_N_DATA_QUESTIONS_PER_WORD
from .going_to_future_bait_content_queries import GET_CONTENT_BAITS
from .lemma_queries import GET_LEMMA_LIST_DATA
from question_builder.data import config
from question_builder.data.domain.data_converters import to_content
from question_builder.data.domain.data_converters import to_verb_question
from question_builder.data.domain.lemma_data import LemmaData
from question_builder.data.domain.content import Content
from neo4j import DirectDriver, Session
from neo4j import BoltStatementResult
from typing import Dict, List


class FDVQuestionGetter(QuestionGetter):
    code = "FDV"
    query_n_data_questions_per_word = GET_N_DATA_QUESTIONS_PER_WORD
    query_n_data_questions = GET_N_DATA_QUESTIONS
    query_lemma_data = GET_LEMMA_LIST_DATA
    query_baits = GET_CONTENT_BAITS

    def get(self, user_id: str, lemma_list: List[str], n_questions: int, n_questions_per_word: int) -> Dict[str, List[LemmaData]]:

        lemma_to_lemma_data_list = super().get(
            user_id, lemma_list, n_questions, n_questions_per_word
        )

        self._append_bait_content(user_id, lemma_to_lemma_data_list)

        return lemma_to_lemma_data_list

    def _append_bait_content(self, user_id: str, lemma_to_lemma_data_list: Dict[str, List[LemmaData]]):
        with self._driver.session() as session:
            for lemma, lemma_data_list in lemma_to_lemma_data_list.items():
                for lemma_data in lemma_data_list:
                    bait_content = self._get_no_going_content(session, user_id, lemma)
                    lemma_data.data_question.bait_content = bait_content

    def _get_no_going_content(self, session: Session, user_id: str, lemma: str) -> Content:

        for res in session.read_transaction(
            self._get_no_going_content_tx, user_id, lemma
        ):
            content_rel_json = res[config.CONTENT_KEY]

            return to_content(content_rel_json[config.CONTENT_KEY])

    @classmethod
    def _get_no_going_content_tx(cls, tx, user_id: str, lemma: str) -> BoltStatementResult:
        return tx.run(cls.query_baits, user_id=user_id, lemma=lemma)



import random

from question_builder.bp.question_creators.question_creator import QuestionCreator
from question_builder.bp.questions.question import Question
from question_builder.data import DataQuestion
from typing import List

FULL_REGULAR_PAST_SUFFIX = "ed"
HALF_REGULAR_PAST_SUFFIX = "d"
Y_REGULAR_PAST_SUFFIX = "ied"
VERB_ENDING_E = "e"
VERB_ENDING_Y = "y"
REMOVE_Y_ENDING = -1


class IrregularPastTenseChoiceQuestionCreator(QuestionCreator):

    CODE = "IRTCP"
    BAITS_CODE = "noirtc-p"

    def create(self, data_question: DataQuestion, user_id: str) -> Question:
        content = data_question.content
        target_lemma = data_question.target_lemma
        target_word = data_question.target_word

        question = Question()
        question.content_id = content.id
        question.target_word = target_word
        question.target_lemma = target_lemma
        question.links, question.media_types = self._get_links_and_media_types(content)
        question.correct_answer = target_word
        question.baits = self._get_bait(target_lemma)
        question.options = self._get_options(question.correct_answer, question.baits)
        question.phrase = self._get_phrase(content.phrase, target_word)
        question.original_phrase = content.phrase
        question.phrase_translation = self._get_translation(content)
        question.question_type = self.CODE
        question.baits_type = self.BAITS_CODE
        return question

    @staticmethod
    def _get_bait(target_lemma: str) -> List[str]:
        if target_lemma.endswith(VERB_ENDING_E):
            return [f"{target_lemma}{HALF_REGULAR_PAST_SUFFIX}"]
        elif target_lemma.endswith(VERB_ENDING_Y):
            target_lemma_no_y = target_lemma[:REMOVE_Y_ENDING]
            return [f"{target_lemma_no_y}{Y_REGULAR_PAST_SUFFIX}"]
        else:
            return [f"{target_lemma}{FULL_REGULAR_PAST_SUFFIX}"]

    def _get_phrase(self, original_phrase: str, target_verb: str) -> str:
        return self._underline_word(original_phrase, target_verb)



from .question_builder import QuestionBuilder
from .question_creators.question_creator import QuestionCreator
from .question_pack import QUESTION_PACK_ALL_CODE
from .question_pack import QUESTION_PACK_TEST_CODE
from .question_pack import question_type_to_creator
from .question_pack import QuestionPack
from .question_pack import CreatorNode
from .questions import question
from .questions.dual_video_question import DualVideoQuestion
from .questions.question import Question



from collections import defaultdict

from . import queries
from typing import Dict
from neo4j import DirectDriver
from neo4j import BoltStatementResult

QUESTION_PACK_KEY = "question_pack"
CODE_KEY = "code"
LEVEL_KEY = "level"
QUESTION_TYPES_KEY = "question_types"
ENABLED_KEY = "enabled"


class QuestionPackRepository:
    """
    Handles all the db operations related to question pack management
    """

    def __init__(self, driver: DirectDriver):
        """
        Initializes the neo4j db
        """
        self._driver = driver

    def get_question_packs(self) -> Dict:
        question_packs = {}
        with self._driver.session() as session:
            for res in session.read_transaction(self._get_question_packs):
                question_pack_code = res[QUESTION_PACK_KEY][CODE_KEY]
                level = res[LEVEL_KEY]
                for question_type in res[QUESTION_TYPES_KEY]:
                    if question_pack_code not in question_packs:
                        question_packs[question_pack_code] = defaultdict(list)
                    if question_type[ENABLED_KEY]:
                        question_packs[question_pack_code][level].append(
                            question_type[CODE_KEY]
                        )
        return question_packs

    @staticmethod
    def _get_question_packs(tx) -> BoltStatementResult:
        query = queries.GET_QUESTION_PACKS
        results = tx.run(query)
        return results



from .question_creator_exception import QuestionCreatorException


class WordNotInAntonymDictionary(QuestionCreatorException):
    pass


class WordNotInEnglishDefinitionsDictionary(QuestionCreatorException):
    pass


class WordNotInSpanishDefinitionsDictionary(QuestionCreatorException):
    pass


class WordNotInSoundsLikeDictionary(QuestionCreatorException):
    pass


class WordNotInIntruderDictionary(QuestionCreatorException):
    pass


class WordNotInMeansLikeDictionary(QuestionCreatorException):
    pass


class WordNotInMultiSynonymDictionary(QuestionCreatorException):
    pass


class WordNotInPartWordTypingDictionary(QuestionCreatorException):
    pass


class WordNotInRelNounDictionary(QuestionCreatorException):
    pass


class WordNotInPreSuffixDictionary(QuestionCreatorException):
    pass


class WordNotInMultisynonymDictionary(QuestionCreatorException):
    pass



import random

from question_builder.bp.question_creators.question_creator import QuestionCreator
from question_builder.bp.questions.question import GETYARN_KEY
from question_builder.bp.questions.question import IMAGE
from question_builder.bp.questions.question import Question
from question_builder.bp.questions.question import TWITTER_KEY
from question_builder.bp.questions.question import VIDEO
from question_builder.data import Content
from typing import List

PROBABILITY_OF_SHUFFLING_OPTIONS_IN_DUAL_VIDEO = 0.5


class MediasList:
    def __init__(self, 
                links: List[str], 
                media_types: List[str], 
                phrases: List[str], 
                correct_values: List[bool]):
        self.links = links
        self.media_types = media_types
        self.phrases = phrases
        self.correct_values = correct_values


class DualVideoQuestionCreator(QuestionCreator):
    def _get_medias(self, contents: List[Content]) -> MediasList:
        """
        Gets links and media types lists depending on type
        """
        links = []
        medias = []
        phrases = []
        correct_values = []
        contents = self._reorder_content(contents)
        for content in contents:
            if content.source == GETYARN_KEY:
                link_mp4 = f"http://media.wordbox.ai/{content.interest_label.lower()}/{content.id}"
                links.append(link_mp4)
                medias.append(VIDEO)
            if content.source == TWITTER_KEY:
                link_jpg = content.image_link
                links.append(link_jpg)
                medias.append(IMAGE)
            phrases.append(content.phrase)
        number_of_contents = len(contents)
        correct_values = [
            False if index != self.correct_option_index else True
            for index in range(0, number_of_contents)
        ]
        return MediasList(links, medias, phrases, correct_values)

    def _reorder_content(self, contents: List[Content]) -> List[Content]:
        correct_option = contents[0]
        if random.random() < PROBABILITY_OF_SHUFFLING_OPTIONS_IN_DUAL_VIDEO:
            random.shuffle(contents)
        self.correct_option_index = contents.index(correct_option)
        return contents
